"""
`analogio` - Analog input and output control
============================================
See `CircuitPython:analogio` in CircuitPython for more details.
Not supported by all boards.

* Author(s): Carter Nelson, Melissa LeBlanc-Williams
"""

import sys

from adafruit_blinka.agnostic import detector

# pylint: disable=ungrouped-imports,wrong-import-position,unused-import

if detector.board.microchip_mcp2221:
    from adafruit_blinka.microcontroller.mcp2221.analogio import AnalogIn
    from adafruit_blinka.microcontroller.mcp2221.analogio import AnalogOut
elif detector.board.greatfet_one:
    from adafruit_blinka.microcontroller.nxp_lpc4330.analogio import AnalogIn
    from adafruit_blinka.microcontroller.nxp_lpc4330.analogio import AnalogOut
elif detector.chip.RK3308:
    from adafruit_blinka.microcontroller.generic_linux.sysfs_analogin import AnalogIn
elif detector.chip.IMX6ULL:
    from adafruit_blinka.microcontroller.generic_linux.sysfs_analogin import AnalogIn
elif "sphinx" in sys.modules:
    pass
else:
    raise NotImplementedError("analogio not supported for this board.")



"""Allows useful indirection to test Pin naming logic by switching platform in testing
    or provide bootstrapping logic for board identification where auto-detection is not
    feasible (e.g. multiple ESP8266 boards architecturally identical). Once runtime
    environment is established, can choose various routes to make available and re-export
    common modules and operations, depending on platform support
"""
import sys
import adafruit_platformdetect

# We intentionally are patching into this namespace as module names so skip the name check.
# pylint: disable=invalid-name

# We'll define board and chip id values in agnostic rather than accessing
# detector directly elsewhere, just in case additional indirection is necessary
# at some later point:

detector = adafruit_platformdetect.Detector()
chip_id = detector.chip.id
board_id = detector.board.id

implementation = sys.implementation.name
if implementation == "micropython":
    from utime import sleep
elif implementation in ("circuitpython", "cpython"):
    from time import sleep



"""Pin definitions for the NanoPi NEO Air."""
# Enable I2C0, UART1, and SPI by adding the following lines to /boot/armbianEnv.txt
#    overlays=usbhost2 usbhost3 spi-spidev uart1 i2c0
#    param_spidev_spi_bus=0

from adafruit_blinka.microcontroller.allwinner.h3 import pin

# Left GPIOs
D2 = pin.PA12
D3 = pin.PA11
D4 = pin.PG11
D17 = pin.PA0
D27 = pin.PA2
D22 = pin.PA3
D10 = pin.PC0
D9 = pin.PC1
D11 = pin.PC2

# Right GPIOs
D14 = pin.PG6
D15 = pin.PG7
D18 = pin.PA6
D23 = pin.PG8
D24 = pin.PG9
D25 = pin.PA1
D8 = pin.PC3

# I2C
SDA = D2
SCL = D3

# SPI
SCLK = D11
MOSI = D10
MISO = D9
CE0 = D8
SCK = SCLK

# Serial UART
UART1_TX = D14
UART1_RX = D15

UART2_RX = D8
UART2_TX = D17
UART2_RTS = D27
UART2_CTS = D22

# PWM
PWM1 = D4



"""A Pin class for use with libgpiod."""
try:
    import gpiod
except ImportError:
    raise ImportError(
        "libgpiod Python bindings not found, please install and try again! See "
        "https://github.com/adafruit/Raspberry-Pi-Installer-Scripts/blob/master/libgpiod.sh"
    ) from ImportError


class Pin:
    """Pins dont exist in CPython so...lets make our own!"""

    IN = 0
    OUT = 1
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2
    _CONSUMER = "adafruit_blinka"

    id = None
    _value = LOW
    _mode = IN

    def __init__(self, pin_id):
        self.id = pin_id
        if isinstance(pin_id, tuple):
            self._num = int(pin_id[1])
            self._chip = gpiod.Chip(str(pin_id[0]), gpiod.Chip.OPEN_BY_NUMBER)
        else:
            self._num = int(pin_id)
            self._chip = gpiod.Chip("gpiochip0", gpiod.Chip.OPEN_BY_NAME)
        self._line = None

    def __repr__(self):
        return str(self.id)

    def __eq__(self, other):
        return self.id == other

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if not self._line:
            self._line = self._chip.get_line(int(self._num))
            # print("init line: ", self.id, self._line)

        if mode is not None:
            if mode == self.IN:
                flags = 0
                if pull is not None:
                    if pull == self.PULL_UP:
                        raise NotImplementedError(
                            "Internal pullups not supported in libgpiod, "
                            "use physical resistor instead!"
                        )
                    if pull == self.PULL_DOWN:
                        raise NotImplementedError(
                            "Internal pulldowns not supported in libgpiod, "
                            "use physical resistor instead!"
                        )
                    raise RuntimeError("Invalid pull for pin: %s" % self.id)

                self._mode = self.IN
                self._line.release()
                self._line.request(
                    consumer=self._CONSUMER, type=gpiod.LINE_REQ_DIR_IN, flags=flags
                )

            elif mode == self.OUT:
                if pull is not None:
                    raise RuntimeError("Cannot set pull resistor on output")
                self._mode = self.OUT
                self._line.release()
                self._line.request(consumer=self._CONSUMER, type=gpiod.LINE_REQ_DIR_OUT)

            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)

    def value(self, val=None):
        """Set or return the Pin Value"""
        if val is None:
            return self._line.get_value()

        if val in (self.LOW, self.HIGH):
            self._value = val
            self._line.set_value(val)
            return None
        raise RuntimeError("Invalid value for pin")



"""Pin definitions for the Clockwork Pi (CPI3) board."""

from adafruit_blinka.microcontroller.allwinner.a33 import pin

# Clockwork Pi GPIO port (DEBUG section in datasheet)

# Type 	Pin # (ext.) 	Pin # (Package) 	Function 1 	Function 2 	Pin # (sysfs) 	Color
# 3V0 	1 											blue
# GPIO 	2 		PB0 			UART0/2_TX 	PB-EINT0 	32 		green
# GPIO 	3 		PB1 			UART0/2_RX 	PB-EINT1 	33 		yellow
# GND 	4 											white
# GPIO 	5 		PH5 			I2C1-SDA 			229 		red
# GPIO 	6 		PH4 			I2C1-SCL 			228 		brown
# GND 	7 											black
# GPIO 	8 		PH6 			UART3-TX 	SPI0-CS 	230 		blue
# GPIO	9 		PH7 			UART3-RX 	SPI0-CLK 	231 		green
# GPIO 	10 		PH9 			UART3-CTS 	SPI0-MISO 	233 		yellow
# GPIO 	11 		PH8 			UART3-RTS 	SPI0-MOSI 	232 		white
# GND 	12 											red
# 5V0 	13 											brown
# 5V0 	14 											black

PB0 = pin.PB0
PB1 = pin.PB1
TX = PB0
RX = PB1

SCL = pin.PH4
SDA = pin.PH5

SCLK = pin.PH7
MOSI = pin.PH8
MISO = pin.PH9



from adafruit_blinka.agnostic import microcontroller

if microcontroller == "esp8266":
    pin_count = 10
elif microcontroller == "samd21":
    pin_count = 38
else:
    raise NotImplementedError("Microcontroller not supported")



"""I2C Class for Binho Nova"""

from adafruit_blinka.microcontroller.nova import Connection


class I2C:
    """Custom I2C Class for Binho Nova"""

    WHR_PAYLOAD_MAX_LENGTH = 1024

    def __init__(self, *, frequency=400000):
        self._nova = Connection.getInstance()
        self._nova.setNumericalBase(10)
        self._nova.setOperationMode(0, "I2C")
        self._nova.setPullUpStateI2C(0, "EN")
        self._nova.setClockI2C(0, frequency)

        self._novaCMDVer = "0"
        if hasattr(self._nova, "getCommandVer"):
            response = self._nova.getCommandVer().split(" ")
            if response[0] != "-NG":
                self._novaCMDVer = response[1]

    def __del__(self):
        """Close Nova on delete"""
        self._nova.close()

    def scan(self):
        """Perform an I2C Device Scan"""
        scanResults = []

        for i in range(8, 121):
            result = self._nova.scanAddrI2C(0, i << 1)

            resp = result.split(" ")

            if resp[3] == "OK":
                scanResults.append(i)

        return scanResults

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        """Write data from the buffer to an address"""
        end = end if end else len(buffer)
        readBytes = 0
        if int(self._novaCMDVer) >= 1:
            chunks, rest = divmod(end - start, self.WHR_PAYLOAD_MAX_LENGTH)
            for i in range(chunks):
                chunk_start = start + i * self.WHR_PAYLOAD_MAX_LENGTH
                chunk_end = chunk_start + self.WHR_PAYLOAD_MAX_LENGTH
                self._nova.writeToReadFromI2C(
                    0,
                    address << 1,
                    stop,
                    readBytes,
                    chunk_end - chunk_start,
                    buffer[chunk_start:chunk_end],
                )
            if rest:
                self._nova.writeToReadFromI2C(
                    0, address << 1, stop, readBytes, rest, buffer[-1 * rest :]
                )
        else:
            self._nova.startI2C(0, address << 1)

            for i in range(start, end):
                self._nova.writeByteI2C(0, buffer[i])

            if stop:
                self._nova.endI2C(0)
            else:
                self._nova.endI2C(0, True)

    # pylint: disable=unused-argument
    def readfrom_into(self, address, buffer, *, start=0, end=None, stop=True):
        """Read data from an address and into the buffer"""
        end = end if end else len(buffer)
        result = self._nova.readBytesI2C(0, address << 1, len(buffer[start:end]))

        if result != "-NG":
            resp = result.split(" ")

            for i in range(len(buffer[start:end])):
                buffer[start + i] = int(resp[2 + i])
        else:
            raise RuntimeError(
                "Received error response from Binho Nova, result = " + result
            )

    # pylint: disable=too-many-locals,too-many-branches
    def writeto_then_readfrom(
        self,
        address,
        buffer_out,
        buffer_in,
        *,
        out_start=0,
        out_end=None,
        in_start=0,
        in_end=None,
        stop=False
    ):
        """Write data from buffer_out to an address and then
        read data from an address and into buffer_in
        """
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)
        if int(self._novaCMDVer) >= 1:
            totalIn = in_end - in_start
            totalOut = out_end - out_start
            totalBytes = totalIn
            if totalOut > totalIn:
                totalBytes = totalOut
            chunks, rest = divmod(totalBytes, self.WHR_PAYLOAD_MAX_LENGTH)
            if rest > 0:
                chunks += 1

            for i in range(chunks):
                # calculate the number of bytes to be written and read
                numInBytes = self.WHR_PAYLOAD_MAX_LENGTH
                if totalIn < self.WHR_PAYLOAD_MAX_LENGTH:
                    numInBytes = totalIn
                numOutBytes = self.WHR_PAYLOAD_MAX_LENGTH
                if totalOut < self.WHR_PAYLOAD_MAX_LENGTH:
                    numOutBytes = totalOut

                # setup the buffer out chunk offset
                buffer = "0"
                if numOutBytes > 0:
                    chunk_start = out_start + i * self.WHR_PAYLOAD_MAX_LENGTH
                    chunk_end = chunk_start + numOutBytes
                    buffer = buffer_out[chunk_start:chunk_end]

                result = self._nova.writeToReadFromI2C(
                    0, address << 1, stop, numInBytes, numOutBytes, buffer
                )
                totalIn -= numInBytes
                totalOut -= numOutBytes

                if result != "-NG":
                    if numInBytes:
                        resp = result.split(" ")
                        resp = resp[2]

                        for j in range(numInBytes):
                            buffer_in[
                                in_start + i * self.WHR_PAYLOAD_MAX_LENGTH + j
                            ] = int(resp[j * 2] + resp[j * 2 + 1], 16)
                else:
                    raise RuntimeError(
                        "Received error response from Binho Nova, result = " + result
                    )
        else:
            self._nova.startI2C(0, address << 1)

            for i in range(out_start, out_end):
                self._nova.writeByteI2C(0, buffer_out[i])

            if stop:
                self._nova.endI2C(0)
            else:
                self._nova.endI2C(0, True)

            result = self._nova.readBytesI2C(
                0, address << 1, len(buffer_in[in_start:in_end])
            )

            if result != "-NG":
                resp = result.split(" ")

                for i in range(len(buffer_in[in_start:in_end])):
                    buffer_in[in_start + i] = int(resp[2 + i])
            else:
                raise RuntimeError(
                    "Received error response from Binho Nova, result = " + result
                )


# pylint: enable=unused-argument,too-many-locals,too-many-branches



"""NXP IMX6ULL pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

# GPIO num = reconment function = Pin((chip, line))
GPIO31 = I2C2_SDA = Pin((0, 31))  # GPIO1_IO31
GPIO30 = I2C2_SCL = Pin((0, 30))  # GPIO1_IO30

GPIO29 = I2C3_SDA = Pin((0, 29))  # GPIO1_IO29
GPIO28 = I2C3_SCL = Pin((0, 28))  # GPIO1_IO28

GPIO24 = UART3_TXD = Pin((0, 24))  # GPIO1_IO24
GPIO25 = UART3_RXD = Pin((0, 25))  # GPIO1_IO25

GPIO22 = ECSPI3_MOSI = Pin((0, 22))  # GPIO1_IO22
GPIO23 = ECSPI3_MISO = Pin((0, 23))  # GPIO1_IO23
GPIO21 = ECSPI3_SCLK = Pin((0, 21))  # GPIO1_IO21
GPIO20 = ECSPI3_SS0 = Pin((0, 20))  # GPIO1_IO20
GPIO18 = ECSPI3_SS1 = Pin((0, 18))  # GPIO1_IO18

GPIO0 = ADC_IN0 = Pin((0, 0))  # GPIO1_IO0
GPIO1 = ADC_IN1 = Pin((0, 1))  # GPIO1_IO2
GPIO2 = ADC_IN2 = Pin((0, 2))  # GPIO1_IO2
GPIO3 = ADC_IN3 = Pin((0, 3))  # GPIO1_IO3
GPIO4 = PWM_C3 = Pin((0, 4))  # GPIO1_IO4
GPIO26 = Pin((0, 26))  # GPIO1_IO26
GPIO27 = Pin((0, 27))  # GPIO1_IO27

GPIO113 = Pin((3, 17))  # GPIO4_IO17
GPIO114 = Pin((3, 18))  # GPIO4_IO18
GPIO115 = PWM_C7 = Pin((3, 19))  # GPIO4_IO19
GPIO116 = PWM_C8 = Pin((3, 20))  # GPIO4_IO20
GPIO117 = Pin((3, 21))  # GPIO4_IO21
GPIO118 = Pin((3, 22))  # GPIO4_IO22
GPIO119 = Pin((3, 23))  # GPIO4_IO23
GPIO120 = Pin((3, 24))  # GPIO4_IO24
GPIO121 = Pin((3, 25))  # GPIO4_IO25
GPIO112 = Pin((3, 26))  # GPIO4_IO26
GPIO123 = Pin((3, 27))  # GPIO4_IO27
GPIO124 = Pin((3, 28))  # GPIO4_IO28

GPIO129 = Pin((4, 1))  # GPIO5_IO1

i2cPorts = (
    (1, I2C2_SCL, I2C2_SDA),
    (2, I2C3_SCL, I2C3_SDA),
)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((2, ECSPI3_SCLK, ECSPI3_MOSI, ECSPI3_MISO),)

# UART3_TXD/RXD on /dev/ttymxc2
uartPorts = ((2, UART3_TXD, UART3_RXD),)

# SysFS pwm outputs, pwm channel and pin in first tuple
pwmOuts = (
    ((2, 0), PWM_C3),
    ((6, 0), PWM_C7),
    ((7, 0), PWM_C8),
)

# SysFS analog inputs, Ordered as analog analogInId, device, and channel
analogIns = (
    (ADC_IN0, 0, 0),
    (ADC_IN1, 0, 1),
    (ADC_IN2, 0, 2),
    (ADC_IN3, 0, 3),
)



"""Generic Connection class for Binho Nova to keep track of connection"""


class Connection:
    """Connection class"""

    __instance = None

    @staticmethod
    def getInstance():
        """ Static access method. """
        if Connection.__instance is None:
            Connection()
        return Connection.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if Connection.__instance is not None:
            raise Exception("This class is a singleton!")

        # pylint: disable=import-outside-toplevel
        from binhoHostAdapter import binhoHostAdapter
        from binhoHostAdapter import binhoUtilities

        # pylint: enable=import-outside-toplevel
        devices = binhoUtilities.listAvailableDevices()

        if len(devices) > 0:
            Connection.__instance = binhoHostAdapter.binhoHostAdapter(devices[0])
        else:
            raise RuntimeError("No Binho Nova found!")



"""
`analogio` - Analog output control
=================================================
Write the SysFS DAC using IIO (Industrial Input/Output)

See `CircuitPython:analogio` in CircuitPython for more details.
* Author(s): Melissa LeBlanc-Williams
"""

import os
from adafruit_blinka import ContextManaged

try:
    from microcontroller.pin import analogOuts
except ImportError:
    raise RuntimeError("No Analog Outputs defined for this board") from ImportError


class AnalogOut(ContextManaged):
    """Analog Output Class"""

    # Sysfs paths
    _sysfs_path = "/sys/bus/iio/devices/"
    _device_path = "iio:device{}"

    # Channel paths
    _channel_path = "out_voltage{}_raw"
    _scale_path = "out_voltage_scale"

    def __init__(self, dac_id):
        """Instantiate an AnalogOut object and verify the sysfs IIO
        corresponding to the specified channel and pin.

        Args:
            dac_id (int): Analog Output ID as defined in microcontroller.pin

        Returns:
            AnalogOut: AnalogOut object.

        Raises:
            TypeError: if `channel` or `pin` types are invalid.
            ValueError: if AnalogOut channel does not exist.

        """

        self.id = dac_id
        self._device = None
        self._channel = None
        self._open(dac_id)

    def __enter__(self):
        return self

    def _open(self, dac_id):
        self._device = None
        for dacpair in analogOuts:
            if dacpair[0] == dac_id:
                self._device = dacpair[1]
                self._channel = dacpair[2]

        if self._device is None:
            raise RuntimeError("No AnalogOut device found for the given ID")

        device_path = os.path.join(
            self._sysfs_path, self._device_path.format(self._device)
        )

        if not os.path.isdir(device_path):
            raise ValueError(
                "AnalogOut device does not exist, check that the required modules are loaded."
            )

    @property
    def value(self):
        """Return an error. This is output only."""
        # emulate what CircuitPython does
        raise AttributeError("unreadable attribute")

    @value.setter
    def value(self, value):
        """Write to the DAC"""
        path = os.path.join(
            self._sysfs_path,
            self._device_path.format(self._device),
            self._channel_path.format(self._channel),
        )

        with open(path, "w") as analog_out:
            return analog_out.write(value + "\n")

    def deinit(self):
        self._device = None
        self._channel = None



"""UART Class for NXP LPC4330"""
from greatfet import GreatFET
from greatfet.interfaces.uart import UART as _UART


class UART:
    """Custom UART Class for NXP LPC4330"""

    PARITY_NONE = 0
    PARITY_ODD = 1
    PARITY_EVEN = 2
    PARITY_STUCK_AT_ONE = 3
    PARITY_STUCK_AT_ZERO = 4

    # pylint: disable=too-many-arguments,unused-argument
    def __init__(
        self,
        portid,
        baudrate=9600,
        bits=8,
        parity=None,
        stop=1,
        timeout=1000,
        read_buf_len=None,
        flow=None,
    ):
        self._gf = GreatFET()
        self._uart = _UART(
            self._gf,
            baud=baudrate,
            data_bits=bits,
            stop_bits=stop,
            parity=parity,
            uart_number=portid,
        )

        if flow is not None:  # default None
            raise NotImplementedError(
                "Parameter '{}' unsupported on GreatFET One".format("flow")
            )

    # pylint: enable=too-many-arguments,unused-argument

    def deinit(self):
        """Deinitialize"""
        self._uart.initialized = False

    def read(self, nbytes=None):
        """Read data from UART and return it"""
        if nbytes is None:
            return None
        return self._uart.read(nbytes)

    def readinto(self, buf, nbytes=None):
        """Read data from UART and into the buffer"""
        if nbytes is None:
            return None
        result = self.read(nbytes)
        for _ in range(nbytes):
            buf.append(result)
        return buf

    def readline(self):
        """Read a single line of data from UART"""
        out = self.read(nbytes=1)
        line = out
        while out != "\r":
            out = self.read(nbytes=1)
            line += out
        return line

    def write(self, buf):
        """Write data from the buffer to UART"""
        return self._uart.write(buf)



"""BCM283x NeoPixel Driver Class"""
import time
import atexit
import _rpi_ws281x as ws

# LED configuration.
# pylint: disable=redefined-outer-name,too-many-branches,too-many-statements
# pylint: disable=global-statement,protected-access
LED_CHANNEL = 0
LED_FREQ_HZ = 800000  # Frequency of the LED signal.  We only support 800KHz
LED_DMA_NUM = 10  # DMA channel to use, can be 0-14.
LED_BRIGHTNESS = 255  # We manage the brightness in the neopixel library
LED_INVERT = 0  # We don't support inverted logic
LED_STRIP = None  # We manage the color order within the neopixel library

# a 'static' object that we will use to manage our PWM DMA channel
# we only support one LED strip per raspi
_led_strip = None
_buf = None


def neopixel_write(gpio, buf):
    """NeoPixel Writing Function"""
    global _led_strip  # we'll have one strip we init if its not at first
    global _buf  # we save a reference to the buf, and if it changes we will cleanup and re-init.

    if _led_strip is None or buf is not _buf:
        # This is safe to call since it doesn't do anything if _led_strip is None
        neopixel_cleanup()

        # Create a ws2811_t structure from the LED configuration.
        # Note that this structure will be created on the heap so you
        # need to be careful that you delete its memory by calling
        # delete_ws2811_t when it's not needed.
        _led_strip = ws.new_ws2811_t()
        _buf = buf

        # Initialize all channels to off
        for channum in range(2):
            channel = ws.ws2811_channel_get(_led_strip, channum)
            ws.ws2811_channel_t_count_set(channel, 0)
            ws.ws2811_channel_t_gpionum_set(channel, 0)
            ws.ws2811_channel_t_invert_set(channel, 0)
            ws.ws2811_channel_t_brightness_set(channel, 0)

        channel = ws.ws2811_channel_get(_led_strip, LED_CHANNEL)

        # Initialize the channel in use
        count = 0
        if len(buf) % 3 == 0:
            # most common, divisible by 3 is likely RGB
            LED_STRIP = ws.WS2811_STRIP_RGB
            count = len(buf) // 3
        elif len(buf) % 4 == 0:
            LED_STRIP = ws.SK6812_STRIP_RGBW
            count = len(buf) // 4
        else:
            raise RuntimeError("We only support 3 or 4 bytes-per-pixel")

        ws.ws2811_channel_t_count_set(
            channel, count
        )  # we manage 4 vs 3 bytes in the library
        ws.ws2811_channel_t_gpionum_set(channel, gpio._pin.id)
        ws.ws2811_channel_t_invert_set(channel, LED_INVERT)
        ws.ws2811_channel_t_brightness_set(channel, LED_BRIGHTNESS)
        ws.ws2811_channel_t_strip_type_set(channel, LED_STRIP)

        # Initialize the controller
        ws.ws2811_t_freq_set(_led_strip, LED_FREQ_HZ)
        ws.ws2811_t_dmanum_set(_led_strip, LED_DMA_NUM)

        resp = ws.ws2811_init(_led_strip)
        if resp != ws.WS2811_SUCCESS:
            if resp == -5:
                raise RuntimeError(
                    "NeoPixel support requires running with sudo, please try again!"
                )
            message = ws.ws2811_get_return_t_str(resp)
            raise RuntimeError(
                "ws2811_init failed with code {0} ({1})".format(resp, message)
            )
        atexit.register(neopixel_cleanup)

    channel = ws.ws2811_channel_get(_led_strip, LED_CHANNEL)
    if gpio._pin.id != ws.ws2811_channel_t_gpionum_get(channel):
        raise RuntimeError("Raspberry Pi neopixel support is for one strip only!")

    if ws.ws2811_channel_t_strip_type_get(channel) == ws.WS2811_STRIP_RGB:
        bpp = 3
    else:
        bpp = 4
    # assign all colors!
    for i in range(len(buf) // bpp):
        r = buf[bpp * i]
        g = buf[bpp * i + 1]
        b = buf[bpp * i + 2]
        if bpp == 3:
            pixel = (r << 16) | (g << 8) | b
        else:
            w = buf[bpp * i + 3]
            pixel = (w << 24) | (r << 16) | (g << 8) | b
        ws.ws2811_led_set(channel, i, pixel)

    resp = ws.ws2811_render(_led_strip)
    if resp != ws.WS2811_SUCCESS:
        message = ws.ws2811_get_return_t_str(resp)
        raise RuntimeError(
            "ws2811_render failed with code {0} ({1})".format(resp, message)
        )
    time.sleep(0.001 * ((len(buf) // 100) + 1))  # about 1ms per 100 bytes


def neopixel_cleanup():
    """Cleanup when we're done"""
    global _led_strip

    if _led_strip is not None:
        # Ensure ws2811_fini is called before the program quits.
        ws.ws2811_fini(_led_strip)
        # Example of calling delete function to clean up structure memory.  Isn't
        # strictly necessary at the end of the program execution here, but is good practice.
        ws.delete_ws2811_t(_led_strip)
        _led_strip = None



"""
`digitalio` - Digital input and output control (GPIO)
=====================================================

See `CircuitPython:digitalio` in CircuitPython for more details.

* Author(s): cefn
"""

from adafruit_blinka.agnostic import board_id, detector

# pylint: disable=ungrouped-imports,wrong-import-position

if detector.chip.BCM2XXX:
    from adafruit_blinka.microcontroller.bcm283x.pin import Pin
elif detector.chip.AM33XX:
    from adafruit_blinka.microcontroller.am335x.pin import Pin
elif detector.chip.DRA74X:
    from adafruit_blinka.microcontroller.dra74x.pin import Pin
elif detector.chip.SUN8I:
    from adafruit_blinka.microcontroller.allwinner.h3.pin import Pin
elif detector.chip.SAMA5:
    from adafruit_blinka.microcontroller.sama5.pin import Pin
elif detector.chip.T210:
    from adafruit_blinka.microcontroller.tegra.t210.pin import Pin
elif detector.chip.T186:
    from adafruit_blinka.microcontroller.tegra.t186.pin import Pin
elif detector.chip.T194:
    from adafruit_blinka.microcontroller.tegra.t194.pin import Pin
elif detector.chip.S905:
    from adafruit_blinka.microcontroller.amlogic.s905.pin import Pin
elif detector.chip.S905X3:
    from adafruit_blinka.microcontroller.amlogic.s905x3.pin import Pin
elif detector.chip.S922X:
    from adafruit_blinka.microcontroller.amlogic.s922x.pin import Pin
elif detector.chip.EXYNOS5422:
    from adafruit_blinka.microcontroller.samsung.exynos5422.pin import Pin
elif detector.chip.APQ8016:
    from adafruit_blinka.microcontroller.snapdragon.apq8016.pin import Pin
elif detector.chip.IMX8MX:
    from adafruit_blinka.microcontroller.nxp_imx8m.pin import Pin
elif detector.chip.IMX6ULL:
    from adafruit_blinka.microcontroller.nxp_imx6ull.pin import Pin
elif detector.chip.HFU540:
    from adafruit_blinka.microcontroller.hfu540.pin import Pin
elif detector.chip.A64:
    from adafruit_blinka.microcontroller.allwinner.a64.pin import Pin
elif detector.chip.A33:
    from adafruit_blinka.microcontroller.allwinner.a33.pin import Pin
elif detector.chip.MIPS24KEC:
    from adafruit_blinka.microcontroller.mips24kec.pin import Pin
elif detector.chip.RK3308:
    from adafruit_blinka.microcontroller.rockchip.rk3308.pin import Pin
elif detector.board.ftdi_ft232h:
    from adafruit_blinka.microcontroller.ft232h.pin import Pin
elif detector.board.binho_nova:
    from adafruit_blinka.microcontroller.nova.pin import Pin
elif detector.board.greatfet_one:
    from adafruit_blinka.microcontroller.nxp_lpc4330.pin import Pin
elif detector.chip.STM32F405:
    from machine import Pin
elif detector.board.microchip_mcp2221:
    from adafruit_blinka.microcontroller.mcp2221.pin import Pin
elif detector.chip.PENTIUM_N3710:
    from adafruit_blinka.microcontroller.pentium.n3710.pin import Pin
elif detector.chip.STM32MP157:
    from adafruit_blinka.microcontroller.stm32.stm32mp157.pin import Pin
elif detector.chip.MT8167:
    from adafruit_blinka.microcontroller.mt8167.pin import Pin
elif detector.chip.H5:
    from adafruit_blinka.microcontroller.allwinner.h5.pin import Pin
elif detector.chip.H616:
    from adafruit_blinka.microcontroller.allwinner.h616.pin import Pin

from adafruit_blinka import Enum, ContextManaged


class DriveMode(Enum):
    """Drive Mode Enumeration"""

    PUSH_PULL = None
    OPEN_DRAIN = None


DriveMode.PUSH_PULL = DriveMode()
DriveMode.OPEN_DRAIN = DriveMode()


class Direction(Enum):
    """Direction Enumeration"""

    INPUT = None
    OUTPUT = None


Direction.INPUT = Direction()
Direction.OUTPUT = Direction()


class Pull(Enum):
    """PullUp/PullDown Enumeration"""

    UP = None
    DOWN = None
    # NONE=None


Pull.UP = Pull()
Pull.DOWN = Pull()

# Pull.NONE = Pull()


class DigitalInOut(ContextManaged):
    """DigitalInOut CircuitPython compatibility implementation"""

    _pin = None

    def __init__(self, pin):
        self._pin = Pin(pin.id)
        self.direction = Direction.INPUT

    def switch_to_output(self, value=False, drive_mode=DriveMode.PUSH_PULL):
        """Switch the Digital Pin Mode to Output"""
        self.direction = Direction.OUTPUT
        self.value = value
        self.drive_mode = drive_mode

    def switch_to_input(self, pull=None):
        """Switch the Digital Pin Mode to Input"""
        self.direction = Direction.INPUT
        self.pull = pull

    def deinit(self):
        """Deinitialize the Digital Pin"""
        del self._pin

    @property
    def direction(self):
        """Get or Set the Digital Pin Direction"""
        return self.__direction

    @direction.setter
    def direction(self, value):
        self.__direction = value
        if value is Direction.OUTPUT:
            self._pin.init(mode=Pin.OUT)
            self.value = False
            self.drive_mode = DriveMode.PUSH_PULL
        elif value is Direction.INPUT:
            self._pin.init(mode=Pin.IN)
            self.pull = None
        else:
            raise AttributeError("Not a Direction")

    @property
    def value(self):
        """Get or Set the Digital Pin Value"""
        return self._pin.value() == 1

    @value.setter
    def value(self, val):
        if self.direction is Direction.OUTPUT:
            self._pin.value(1 if val else 0)
        else:
            raise AttributeError("Not an output")

    @property
    def pull(self):
        """Get or Set the Digital Pin Direction"""
        if self.direction is Direction.INPUT:
            return self.__pull
        raise AttributeError("Not an input")

    @pull.setter
    def pull(self, pul):
        if self.direction is Direction.INPUT:
            self.__pull = pul
            if pul is Pull.UP:
                self._pin.init(mode=Pin.IN, pull=Pin.PULL_UP)
            elif pul is Pull.DOWN:
                if hasattr(Pin, "PULL_DOWN"):
                    self._pin.init(mode=Pin.IN, pull=Pin.PULL_DOWN)
                else:
                    raise NotImplementedError(
                        "{} unsupported on {}".format(Pull.DOWN, board_id)
                    )
            elif pul is None:
                self._pin.init(mode=Pin.IN, pull=None)
            else:
                raise AttributeError("Not a Pull")
        else:
            raise AttributeError("Not an input")

    @property
    def drive_mode(self):
        """Get or Set the Digital Pin Drive Mode"""
        if self.direction is Direction.OUTPUT:
            return self.__drive_mode  #
        raise AttributeError("Not an output")

    @drive_mode.setter
    def drive_mode(self, mod):
        self.__drive_mode = mod
        if mod is DriveMode.OPEN_DRAIN:
            self._pin.init(mode=Pin.OPEN_DRAIN)
        elif mod is DriveMode.PUSH_PULL:
            self._pin.init(mode=Pin.OUT)



"""MCP2221 pin names"""
from .mcp2221 import mcp2221


class Pin:
    """A basic Pin class for use with MCP2221."""

    # pin modes
    OUT = 0
    IN = 1
    ADC = 2
    DAC = 3
    # pin values
    LOW = 0
    HIGH = 1

    def __init__(self, pin_id=None):
        self.id = pin_id
        self._mode = None

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if self.id is None:
            raise RuntimeError("Can not init a None type pin.")
        if pull is not None:
            raise NotImplementedError("Internal pullups and pulldowns not supported")
        if mode in (Pin.IN, Pin.OUT):
            # All pins can do GPIO
            mcp2221.gp_set_mode(self.id, mcp2221.GP_GPIO)
            mcp2221.gpio_set_direction(self.id, mode)
        elif mode == Pin.ADC:
            # ADC only available on these pins
            if self.id not in (1, 2, 3):
                raise ValueError("Pin does not have ADC capabilities")
            mcp2221.gp_set_mode(self.id, mcp2221.GP_ALT0)
            mcp2221.adc_configure()
        elif mode == Pin.DAC:
            # DAC only available on these pins
            if self.id not in (2, 3):
                raise ValueError("Pin does not have DAC capabilities")
            mcp2221.gp_set_mode(self.id, mcp2221.GP_ALT1)
            mcp2221.dac_configure()
        else:
            raise ValueError("Incorrect pin mode: {}".format(mode))
        self._mode = mode

    def value(self, val=None):
        """Set or return the Pin Value"""
        # Digital In / Out
        if self._mode in (Pin.IN, Pin.OUT):
            # digital read
            if val is None:
                return mcp2221.gpio_get_pin(self.id)
            # digital write
            if val in (Pin.LOW, Pin.HIGH):
                mcp2221.gpio_set_pin(self.id, val)
                return None
            # nope
            raise ValueError("Invalid value for pin.")
        # Analog In
        if self._mode == Pin.ADC:
            if val is None:
                # MCP2221 ADC is 10 bit, scale to 16 bit per CP API
                return mcp2221.adc_read(self.id) * 64
            # read only
            raise AttributeError("'AnalogIn' object has no attribute 'value'")
        # Analog Out
        if self._mode == Pin.DAC:
            if val is None:
                # write only
                raise AttributeError("unreadable attribute")
            # scale 16 bit value to MCP2221 5 bit DAC (yes 5 bit)
            mcp2221.dac_write(self.id, val // 2048)
            return None
        raise RuntimeError(
            "No action for mode {} with value {}".format(self._mode, val)
        )


# create pin instances for each pin
G0 = Pin(0)
G1 = Pin(1)
G2 = Pin(2)
G3 = Pin(3)

SCL = Pin()
SDA = Pin()



import gc
from unittest import TestCase
from testing import await_true

gc.collect()


class TestGPSInteractive(TestCase):
    def test_read_value(self):
        import adafruit_blinka

        adafruit_blinka.patch_system()  # needed before adafruit_gps imports time

        import microcontroller.pin

        gc.collect()
        import busio

        gc.collect()
        import adafruit_gps

        gc.collect()

        # configure the last available UART (first uart often for REPL)
        uartId, uartTx, uartRx = microcontroller.pin.uartPorts[0]
        uart = busio.UART(uartTx, uartRx, baudrate=9600, timeout=3000)

        gps = adafruit_gps.GPS(uart)

        gps.send_command("PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0")
        gps.send_command("PMTK220,1000")

        def try_fix():
            gps.update()
            return gps.has_fix

        await_true("GPS fix", try_fix)

        self.assertTrue(gps.satellites is not None)
        self.assertTrue(-90 <= gps.latitude < 90)
        self.assertTrue(-180 <= gps.longitude < 180)



"""
Samsum Exynos5422
Ref:
    Linux kernel 4.14.y (hardkernel)
"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

GPIOA0_0 = GPIO171 = Pin((26, 0))
GPIOA0_1 = GPIO172 = Pin((26, 1))
GPIOA0_2 = GPIO173 = Pin((26, 2))
GPIOA0_3 = GPIO174 = Pin((26, 3))
GPIOA0_4 = GPIO175 = Pin((26, 4))
GPIOA0_5 = GPIO176 = Pin((26, 5))
GPIOA0_6 = GPIO177 = Pin((26, 6))
GPIOA0_6 = GPIO178 = Pin((26, 7))

GPIOA2_0 = GPIO185 = Pin((28, 0))
GPIOA2_1 = GPIO186 = Pin((28, 1))
GPIOA2_2 = GPIO187 = Pin((28, 2))
GPIOA2_3 = GPIO188 = Pin((28, 3))
GPIOA2_4 = GPIO189 = Pin((28, 4))
GPIOA2_5 = GPIO190 = Pin((28, 5))
GPIOA2_6 = GPIO191 = Pin((28, 6))
GPIOA2_7 = GPIO192 = Pin((28, 7))

GPIOB3_0 = GPIO207 = Pin((32, 0))
GPIOB3_1 = GPIO208 = Pin((32, 1))
GPIOB3_2 = GPIO209 = Pin((32, 2))
GPIOB3_3 = GPIO210 = Pin((32, 3))
GPIOB3_4 = GPIO211 = Pin((32, 4))
GPIOB3_5 = GPIO212 = Pin((32, 5))
GPIOB3_6 = GPIO213 = Pin((32, 6))
GPIOB3_7 = GPIO214 = Pin((32, 7))

GPIOX1_0 = GPIO16 = Pin((2, 0))
GPIOX1_1 = GPIO17 = Pin((2, 1))
GPIOX1_2 = GPIO18 = Pin((2, 2))
GPIOX1_3 = GPIO19 = Pin((2, 3))
GPIOX1_4 = GPIO20 = Pin((2, 4))
GPIOX1_5 = GPIO21 = Pin((2, 5))
GPIOX1_6 = GPIO22 = Pin((2, 6))
GPIOX1_7 = GPIO23 = Pin((2, 7))

GPIOX2_0 = GPIO24 = Pin((3, 0))
GPIOX2_1 = GPIO25 = Pin((3, 1))
GPIOX2_2 = GPIO26 = Pin((3, 2))
GPIOX2_3 = GPIO27 = Pin((3, 3))
GPIOX2_4 = GPIO28 = Pin((3, 4))
GPIOX2_5 = GPIO29 = Pin((3, 5))
GPIOX2_6 = GPIO30 = Pin((3, 6))
GPIOX2_7 = GPIO31 = Pin((3, 7))

GPIOX3_0 = GPIO32 = Pin((4, 0))
GPIOX3_1 = GPIO33 = Pin((4, 1))
GPIOX3_2 = GPIO34 = Pin((4, 2))
GPIOX3_3 = GPIO35 = Pin((4, 3))
GPIOX3_4 = GPIO36 = Pin((4, 4))
GPIOX3_5 = GPIO37 = Pin((4, 5))
GPIOX3_6 = GPIO38 = Pin((4, 6))
GPIOX3_7 = GPIO39 = Pin((4, 7))

I2C1_SDA = GPIOB3_2
I2C1_SCL = GPIOB3_3
I2C5_SDA = GPIOA2_2
I2C5_SCL = GPIOA2_3

UART0_TX = GPIOA0_1
UART0_RX = GPIOA0_0

SPI1_SCLK = GPIOA2_4
SPI1_MISO = GPIOA2_6
SPI1_MOSI = GPIOA2_7
SPI1_CS0 = GPIOA2_5

# ordered as i2cId, sclId, sdaId
i2cPorts = (
    (1, I2C1_SCL, I2C1_SDA),
    (5, I2C5_SCL, I2C5_SDA),
)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((1, SPI1_SCLK, SPI1_MOSI, SPI1_MISO),)

# ordered as uartId, txId, rxId
uartPorts = ((0, UART0_TX, UART0_RX),)



# mitigate heap fragmentation issues by pre-loading major libraries
import gc

gc.collect()
import unittest

gc.collect()


def yes_no(q, default=True):
    a = input(q + " (Y/n)?" if default else " (y/N)?")
    a = a.lower()
    if a == "":
        return default
    elif a == "n":
        a = False
    elif a == "y":
        a = True
    return a


def multi_choice(q, choices, defaultPos=None):
    if defaultPos is not None:
        print("{} [{}]?".format(q, defaultPos))
    else:
        print(q + "?")
    for pos, choice in enumerate(choices):
        print("{}) {}".format(pos, choice))
    a = input()
    a = a.lower()
    try:
        if a == "":
            a = defaultPos
        else:
            a = int(a)
        return choices[a]
    except Exception as e:
        print(e)
        return None


def await_true(name, fun, interval=0, patience=60):
    from adafruit_blinka.agnostic.time import sleep, monotonic

    print("Waiting {} sec until {} (CTRL+C give up)".format(patience, name))

    deadline = monotonic() + patience
    try:
        while deadline - monotonic() > 0:
            if fun():
                return True
            else:
                sleep(interval)
        return False
    except KeyboardInterrupt:
        return False


def test_module(module, runner=None):
    import unittest

    if runner is None:
        runner = unittest.TestRunner()
    suite = unittest.TestSuite()
    for key in dir(module):
        val = getattr(module, key)
        try:
            if issubclass(val, unittest.TestCase):
                suite.addTest(val)
        except:
            pass
    return runner.run(suite)


def test_module_name(absolute, runner=None):
    try:
        print("Suite begin: {}".format(absolute))
        module = __import__(absolute)
        relatives = absolute.split(".")
        if len(relatives) > 1:
            for relative in relatives[1:]:
                module = getattr(module, relative)
        return test_module(module, runner)
    finally:
        print("Suite end: {}".format(absolute))


def test_interactive(*module_names):
    for module_name in module_names:
        if yes_no("Run suite {}".format(module_name)):
            gc.collect()
            test_module_name(module_name)


def test_prepare(casetype):
    case = casetype()
    case.setUp()


def main():
    """
    moduleNames = ["testing.implementation.universal.digitalio",]
    if agnostic.implementation == "micropython":
        moduleNames.extend([ "testing.implementation.micropython.digitalio",])

    """
    moduleNames = ["testing.implementation.universal.bitbangio"]

    unittest.raiseException = (
        True  # terminates with stack information on userspace Exception
    )
    unittest.raiseBaseException = (
        True  # terminates with stack information on system Exception
    )
    test_interactive(*moduleNames)


gc.collect()



"""Pin definitions for the Rock Pi S."""

from adafruit_blinka.microcontroller.rockchip.rk3308 import pin

D2 = pin.GPIO0_B3
D3 = pin.GPIO0_B4
D4 = pin.GPIO2_A4
D8 = pin.GPIO1_D1
D9 = pin.GPIO1_C6
D10 = pin.GPIO1_C7
D11 = pin.GPIO1_D0
D14 = pin.GPIO2_A1
D15 = pin.GPIO2_A0
D17 = pin.GPIO0_B7
D18 = pin.GPIO2_A5
D22 = pin.GPIO0_C1
D23 = pin.GPIO2_B2
D24 = pin.GPIO2_B1
D25 = pin.GPIO2_A7
D27 = pin.GPIO0_C0

SDA0 = pin.I2C0_SDA
SCL0 = pin.I2C0_SCL

SDA1 = pin.I2C1_SDA
SCL1 = pin.I2C1_SCL

SCL2 = pin.I2C2_SCL
SDA2 = pin.I2C2_SDA

SCL3 = pin.I2C3_SCL
SDA3 = pin.I2C3_SDA

SDA = SDA1
SCL = SCL1

SCLK = pin.SPI2_SCLK
MOSI = pin.SPI2_MOSI
MISO = pin.SPI2_MISO
CS = pin.SPI2_CS
SCK = SCLK

UART_TX = pin.UART0_TX
UART_RX = pin.UART0_RX

UART1_TX = pin.UART1_TX
UART1_RX = pin.UART1_RX

UART2_TX = pin.UART2_TX
UART2_RX = pin.UART2_RX

PWM2 = pin.PWM2
PWM3 = pin.PWM3

ADC_IN0 = pin.ADC_IN0



import time
import sys
import board
import busio

print("hello blinka!")

i2c = busio.I2C(board.SCL, board.SDA)

print("I2C devices found: ", [hex(i) for i in i2c.scan()])

if not 0x18 in i2c.scan():
    print("Didn't find MCP9808")
    sys.exit()


def temp_c(data):
    value = data[0] << 8 | data[1]
    temp = (value & 0xFFF) / 16.0
    if value & 0x1000:
        temp -= 256.0
    return temp


while True:
    i2c.writeto(0x18, bytes([0x05]), stop=False)
    result = bytearray(2)
    i2c.readfrom_into(0x18, result)
    print(temp_c(result))
    time.sleep(0.5)



"""Pin definitions for the Binho Nova"""
from adafruit_blinka.microcontroller.nova import pin

IO0 = pin.IO0
IO1 = pin.IO1
IO2 = pin.IO2
IO3 = pin.IO3
IO4 = pin.IO4

TX = IO4
RX = IO3

SDA = pin.SDA
SCL = pin.SCL

SCK = pin.SCK
SCLK = pin.SCLK
MOSI = pin.MOSI
MISO = pin.MISO
SS0 = pin.SS0
SS1 = pin.SS1



"""Pin definitions for the Beaglebone Black."""
from adafruit_blinka.microcontroller.dra74x import pin

# initial pins, to mimic bonescript demo
# BeagleBone Black
# P8_1 = DGND        # DGND
# P8_2 = DGND        # DGND
P8_3 = pin.P8_3  # GPIO1_6 - GPIO_38
P8_4 = pin.P8_4  # GPIO1_7 - GPIO_39
P8_5 = pin.P8_5  # GPIO1_2 - GPIO_34
P8_6 = pin.P8_6  # GPIO1_3 - GPIO_35
P8_7 = pin.P8_7  # TIMER4 - GPIO_66
P8_8 = pin.P8_8  # TIMER7 - GPIO_67
P8_9 = pin.P8_9  # TIMER5 - GPIO_69
P8_10 = pin.P8_10  # TIMER6 - GPIO_68
P8_11 = pin.P8_11  # GPIO1_13 - GPIO_45
P8_12 = pin.P8_12  # GPIO1_12 - GPIO_44
P8_13 = pin.P8_13  # EHRPWM2B - GPIO_23
P8_14 = pin.P8_14  # GPIO0_26 - GPIO_26
P8_15 = pin.P8_15  # GPIO1_15 - GPIO_47
P8_16 = pin.P8_16  # GPIO1_14 - GPIO_46
P8_17 = pin.P8_17  # GPIO0_27 - GPIO_27
P8_18 = pin.P8_18  # GPIO2_1 - GPIO_65
P8_19 = pin.P8_19  # EHRPWM2A - GPIO_22
P8_20 = pin.P8_20  # GPIO1_31 - GPIO_63
P8_21 = pin.P8_21  # GPIO1_30 - GPIO_62
P8_22 = pin.P8_22  # GPIO1_5 - GPIO_37
P8_23 = pin.P8_23  # GPIO1_4 - GPIO_36
P8_24 = pin.P8_24  # GPIO1_1 - GPIO_33
P8_25 = pin.P8_25  # GPIO1_0 - GPIO_32
P8_26 = pin.P8_26  # GPIO1_29 - GPIO_61
P8_27 = pin.P8_27  # GPIO2_22 - GPIO_86
P8_28 = pin.P8_28  # GPIO2_24 - GPIO_88
P8_29 = pin.P8_29  # GPIO2_23 - GPIO_87
P8_30 = pin.P8_30  # GPIO2_25 - GPIO_89
P8_31 = pin.P8_31  # UART5_CTSN - GPIO_10
P8_32 = pin.P8_32  # UART5_RTSN - GPIO_11
P8_33 = pin.P8_33  # UART4_RTSN - GPIO_9
P8_34 = pin.P8_34  # UART3_RTSN - GPIO_81
P8_35 = pin.P8_35  # UART4_CTSN - GPIO_8
P8_36 = pin.P8_36  # UART3_CTSN - GPIO_80
P8_37 = pin.P8_37  # UART5_TXD - GPIO_78
P8_38 = pin.P8_38  # UART5_RXD - GPIO_79
P8_39 = pin.P8_39  # GPIO2_12 - GPIO_76
P8_40 = pin.P8_40  # GPIO2_13 - GPIO_77
P8_41 = pin.P8_41  # GPIO2_10 - GPIO_74
P8_42 = pin.P8_42  # GPIO2_11 - GPIO_75
P8_43 = pin.P8_43  # GPIO2_8 - GPIO_72
P8_44 = pin.P8_44  # GPIO2_9 - GPIO_73
P8_45 = pin.P8_45  # GPIO2_6 - GPIO_70
P8_46 = pin.P8_46  # GPIO2_7 - GPIO_71

# P9_1 = DGND        # DGND
# P9_2 = DGND        # DGND
# P9_3 = VDD_3V3     # VDD_3V3
# P9_4 = VDD_3V3     # VDD_3V3
# P9_5 = VDD_5V      # VDD_5V
# P9_6 = VDD_5V      # VDD_5V
# P9_7 = SYS_5V      # SYS_5V
# P9_8 = SYS_5V      # SYS_5V
# P9_9 = PWR_BUT     # PWR_BUT
# P9_10 = SYS_RESETN # SYS_RESETn
P9_11 = pin.P9_11  # UART4_RXD - GPIO_30
P9_12 = pin.P9_12  # GPIO1_28 - GPIO_60
P9_13 = pin.P9_13  # UART4_TXD - GPIO_31
P9_14 = pin.P9_14  # EHRPWM1A - GPIO_50
P9_15 = pin.P9_15  # GPIO1_16 - GPIO_48
P9_16 = pin.P9_16  # EHRPWM1B - GPIO_51
P9_17 = pin.P9_17  # I2C1_SCL - GPIO_5
P9_18 = pin.P9_18  # I2C1_SDA - GPIO_4
P9_19 = pin.P9_19  # I2C2_SCL - GPIO_13
P9_20 = pin.P9_20  # I2C2_SDA - GPIO_12
P9_21 = pin.P9_21  # UART2_TXD - GPIO_3
P9_22 = pin.P9_22  # UART2_RXD - GPIO_2
P9_23 = pin.P9_23  # GPIO1_17 - GPIO_49
P9_24 = pin.P9_24  # UART1_TXD - GPIO_15
P9_25 = pin.P9_25  # GPIO3_21 - GPIO_117
P9_26 = pin.P9_26  # UART1_RXD - GPIO_14
P9_27 = pin.P9_27  # GPIO3_19 - GPIO_115
P9_28 = pin.P9_28  # SPI1_CS0 - GPIO_113
P9_29 = pin.P9_29  # SPI1_D0 - GPIO_111
P9_30 = pin.P9_30  # SPI1_D1 - GPIO_112
P9_31 = pin.P9_31  # SPI1_SCLK - GPIO_110
# P9_32 = VDD_ADC    # VDD_ADC
# P9_33 = AIN4       # AIN4
# P9_34 = GNDA_ADC   # GNDA_ADC
# P9_35 = AIN6       # AIN6
# P9_36 = AIN5       # AIN5
# P9_37 = AIN2       # AIN2
# P9_38 = AIN3       # AIN3
# P9_39 = AIN0       # AIN0
# P9_40 = AIN1       # AIN1
P9_41 = pin.P9_41  # CLKOUT2 - GPIO_20
P9_42 = pin.P9_42  # GPIO0_7 - GPIO_7
# P9_43 = DGND       # DGND
# P9_44 = DGND       # DGND
# P9_45 = DGND       # DGND
# P9_46 = DGND       # DGND

# common to all beagles
LED_USR0 = pin.USR0
LED_USR1 = pin.USR1
LED_USR2 = pin.USR2
LED_USR3 = pin.USR3
LED_USR4 = pin.USR4

# I2C and SPI pins from:
# src/adafruit_blinka/board/raspi_40pin.py
# SDA = pin.SDA
# SCL = pin.SCL
# CE1 = pin.D7
# CE0 = pin.D8
# MISO = pin.D9
# MOSI = pin.D10
# SCLK = pin.D11
# SCK = pin.D11
# TXD = pin.D14
# RXD = pin.D15
# MISO_1 = pin.D19
# MOSI_1 = pin.D20
# SCLK_1 = pin.D21
# SCK_1 = pin.D21

SDA = pin.I2C4_SDA  # P9_19
SCL = pin.I2C4_SCL  # P9_20

# Pins for SPI
#
# To enable SPI and an additional I2C port, add the following line to /boot/uEnv.txt:
#    dtb=am5729-beagleboneai-roboticscape.dtb
#
# You can verify the dtb file exists by checking the /boot/dtbs/{kernel_version}/ folder
#
CE0 = pin.SPI1_CS0  # P9_28
CE1 = pin.SPI1_CS1  # P9_42
MOSI = pin.SPI1_D1  # P9_29
MISO = pin.SPI1_D0  # P9_30
SCLK = pin.SPI1_SCLK  # P9_31
# CircuitPython naming convention for SPI Clock
SCK = SCLK



"""PyBoard pin names"""

from adafruit_blinka.microcontroller.stm32.stm32f405 import pin

X1 = pin.A0
X2 = pin.A1
X3 = pin.A2
X4 = pin.A3
X5 = pin.A4
X6 = pin.A5
X7 = pin.A6
X8 = pin.A7
X9 = pin.B6
X10 = pin.B7
X11 = pin.C4
X12 = pin.C5
X17 = pin.B3
X18 = pin.C13
X19 = pin.C0
X20 = pin.C1
X21 = pin.C2
X22 = pin.C3
Y1 = pin.C6
Y2 = pin.C7
Y3 = pin.B8
Y4 = pin.B9
Y5 = pin.B12
Y6 = pin.B13
Y7 = pin.B14
Y8 = pin.B15
Y9 = pin.B10
Y10 = pin.B11
Y11 = pin.B0
Y12 = pin.B1
SW = pin.B3
LED_RED = pin.A13
LED_GREEN = pin.A14
LED_YELLOW = pin.A15
LED_BLUE = pin.B4
MMA_INT = pin.B2
MMA_AVDD = pin.B5
SD_D0 = pin.C8
SD_D1 = pin.C9
SD_D2 = pin.C10
SD_D3 = pin.C11
SD_CMD = pin.D2
SD_CK = pin.C12
SD = pin.A8
SD_SW = pin.A8
USB_VBUS = pin.A9
USB_ID = pin.A10
USB_DM = pin.A11
USB_DP = pin.A12



"""UART Class for Binho Nova"""

from adafruit_blinka.microcontroller.nova import Connection


class UART:
    """Custom UART Class for Binho Nova"""

    ESCAPE_SEQUENCE = "+++UART0"

    # pylint: disable=too-many-arguments,unused-argument
    def __init__(
        self,
        portid,
        baudrate=9600,
        bits=8,
        parity=None,
        stop=1,
        timeout=1000,
        read_buf_len=None,
        flow=None,
    ):
        self._nova = Connection.getInstance()

        self._id = portid
        self._baudrate = baudrate
        self._parity = parity
        self._bits = bits
        self._stop = stop
        self._timeout = timeout

        if flow is not None:  # default 0
            raise NotImplementedError(
                "Parameter '{}' unsupported on Binho Nova".format("flow")
            )

        self._nova.setOperationMode(self._id, "UART")
        self._nova.setBaudRateUART(self._id, baudrate)
        self._nova.setDataBitsUART(self._id, bits)
        self._nova.setParityUART(self._id, parity)
        self._nova.setStopBitsUART(self._id, stop)
        self._nova.setEscapeSequenceUART(self._id, UART.ESCAPE_SEQUENCE)
        self._nova.beginBridgeUART(self._id)

    # pylint: enable=too-many-arguments,unused-argument
    def __del__(self):
        """Close Nova on delete"""
        self.deinit()
        self._nova.close()

    def deinit(self):
        """Deinitialize"""
        self._nova.writeBridgeUART(UART.ESCAPE_SEQUENCE)
        self._nova.stopBridgeUART(UART.ESCAPE_SEQUENCE)

    def read(self, nbytes=None):
        """Read data from UART and return it"""
        if nbytes is None:
            return None
        data = bytearray()
        for _ in range(nbytes):
            data.append(ord(self._nova.readBridgeUART()))
        return data

    def readinto(self, buf, nbytes=None):
        """Read data from UART and into the buffer"""
        if nbytes is None:
            return None
        for _ in range(nbytes):
            buf.append(ord(self._nova.readBridgeUART()))
        return buf

    def readline(self):
        """Read a single line of data from UART"""
        out = self._nova.readBridgeUART()
        line = out
        while out != "\r":
            out = self._nova.readBridgeUART()
            line += out
        return line

    def write(self, buf):
        """Write data from the buffer to UART"""
        return self._nova.writeBridgeUART(buf)



"""
G12A, G12B, and SM1 Common Definitions
Ref:
Linux kernel 4.9.y (hardkernel)
    linux/include/dt-bindings/gpio/meson-g12a-gpio.h
Linux kernel 5.4.y (mainline)
    linux/include/dt-bindings/gpio/meson-g12a-gpio.h
    linux/arch/arm64/boot/dts/amlogic/meson-g12-common.dtsi
"""

import re
import gpiod
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

chip0 = gpiod.Chip("0")
chip1 = gpiod.Chip("1")

if chip0.num_lines() < 20:
    aobus = 0
    periphs = 1
    periphs_offset = chip1.num_lines() - 85
else:
    aobus = 1
    periphs = 0
    periphs_offset = chip0.num_lines() - 85

chip0.close()
chip1.close()

GPIOAO_0 = GPIO496 = Pin((aobus, 0))
GPIOAO_1 = GPIO497 = Pin((aobus, 1))
GPIOAO_2 = GPIO498 = Pin((aobus, 2))
GPIOAO_3 = GPIO499 = Pin((aobus, 3))
GPIOAO_4 = GPIO500 = Pin((aobus, 4))
GPIOAO_5 = GPIO501 = Pin((aobus, 5))
GPIOAO_6 = GPIO502 = Pin((aobus, 6))
GPIOAO_7 = GPIO503 = Pin((aobus, 7))
GPIOAO_8 = GPIO504 = Pin((aobus, 8))
GPIOAO_9 = GPIO505 = Pin((aobus, 9))
GPIOAO_10 = GPIO506 = Pin((aobus, 10))
GPIOAO_11 = GPIO507 = Pin((aobus, 11))
GPIOE_0 = GPIO508 = Pin((aobus, 12))
GPIOE_1 = GPIO509 = Pin((aobus, 13))
GPIOE_2 = GPIO510 = Pin((aobus, 14))
GPIO_TEST_N = GPIO511 = Pin((aobus, 15))

GPIOH_0 = GPIO427 = Pin((periphs, 16 + periphs_offset))
GPIOH_1 = GPIO428 = Pin((periphs, 17 + periphs_offset))
GPIOH_2 = GPIO429 = Pin((periphs, 18 + periphs_offset))
GPIOH_3 = GPIO430 = Pin((periphs, 19 + periphs_offset))
GPIOH_4 = GPIO431 = Pin((periphs, 20 + periphs_offset))
GPIOH_5 = GPIO432 = Pin((periphs, 21 + periphs_offset))
GPIOH_6 = GPIO433 = Pin((periphs, 22 + periphs_offset))
GPIOH_7 = GPIO434 = Pin((periphs, 23 + periphs_offset))
GPIOH_8 = GPIO435 = Pin((periphs, 24 + periphs_offset))

GPIOA_0 = GPIO460 = Pin((periphs, 49 + periphs_offset))
GPIOA_1 = GPIO461 = Pin((periphs, 50 + periphs_offset))
GPIOA_2 = GPIO462 = Pin((periphs, 51 + periphs_offset))
GPIOA_3 = GPIO463 = Pin((periphs, 52 + periphs_offset))
GPIOA_4 = GPIO464 = Pin((periphs, 53 + periphs_offset))
GPIOA_5 = GPIO465 = Pin((periphs, 54 + periphs_offset))
GPIOA_6 = GPIO466 = Pin((periphs, 55 + periphs_offset))
GPIOA_7 = GPIO467 = Pin((periphs, 56 + periphs_offset))
GPIOA_8 = GPIO468 = Pin((periphs, 57 + periphs_offset))
GPIOA_9 = GPIO469 = Pin((periphs, 58 + periphs_offset))
GPIOA_10 = GPIO470 = Pin((periphs, 59 + periphs_offset))
GPIOA_11 = GPIO471 = Pin((periphs, 60 + periphs_offset))
GPIOA_12 = GPIO472 = Pin((periphs, 61 + periphs_offset))
GPIOA_13 = GPIO473 = Pin((periphs, 62 + periphs_offset))
GPIOA_14 = GPIO474 = Pin((periphs, 63 + periphs_offset))
GPIOA_15 = GPIO475 = Pin((periphs, 64 + periphs_offset))

GPIOX_0 = GPIO476 = Pin((periphs, 65 + periphs_offset))
GPIOX_1 = GPIO477 = Pin((periphs, 66 + periphs_offset))
GPIOX_2 = GPIO478 = Pin((periphs, 67 + periphs_offset))
GPIOX_3 = GPIO479 = Pin((periphs, 68 + periphs_offset))
GPIOX_4 = GPIO480 = Pin((periphs, 69 + periphs_offset))
GPIOX_5 = GPIO481 = Pin((periphs, 70 + periphs_offset))
GPIOX_6 = GPIO482 = Pin((periphs, 71 + periphs_offset))
GPIOX_7 = GPIO483 = Pin((periphs, 72 + periphs_offset))
GPIOX_8 = GPIO484 = Pin((periphs, 73 + periphs_offset))
GPIOX_9 = GPIO485 = Pin((periphs, 74 + periphs_offset))
GPIOX_10 = GPIO486 = Pin((periphs, 75 + periphs_offset))
GPIOX_11 = GPIO487 = Pin((periphs, 76 + periphs_offset))
GPIOX_12 = GPIO488 = Pin((periphs, 77 + periphs_offset))
GPIOX_13 = GPIO489 = Pin((periphs, 78 + periphs_offset))
GPIOX_14 = GPIO490 = Pin((periphs, 79 + periphs_offset))
GPIOX_15 = GPIO491 = Pin((periphs, 80 + periphs_offset))
GPIOX_16 = GPIO492 = Pin((periphs, 81 + periphs_offset))
GPIOX_17 = GPIO493 = Pin((periphs, 82 + periphs_offset))
GPIOX_18 = GPIO494 = Pin((periphs, 83 + periphs_offset))
GPIOX_19 = GPIO495 = Pin((periphs, 84 + periphs_offset))

SPI0_SCLK = GPIOX_11
SPI0_MISO = GPIOX_9
SPI0_MOSI = GPIOX_8
SPI0_CS0 = GPIOX_10

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, SPI0_SCLK, SPI0_MOSI, SPI0_MISO),)

UART1_TX = GPIOX_12
UART1_RX = GPIOX_13

# ordered as uartId, txId, rxId
uartPorts = ((1, UART1_TX, UART1_RX),)


def get_dts_alias(device: str) -> str:
    """Get the Device Tree Alias"""
    uevent_path = "/sys/bus/platform/devices/" + device + "/uevent"
    with open(uevent_path, "r") as fd:
        pattern = r"^OF_ALIAS_0=(.*)$"
        uevent = fd.read().split("\n")
        for line in uevent:
            match = re.search(pattern, line)
            if match:
                return match.group(1).upper()

        return None


# ordered as i2cId, sclId, sdaId
i2cPorts = []

alias = get_dts_alias("ffd1d000.i2c")
if alias is not None:
    globals()[alias + "_SCL"] = GPIOX_18
    globals()[alias + "_SDA"] = GPIOX_17
    i2cPorts.append((int(alias[3]), GPIOX_18, GPIOX_17))

alias = get_dts_alias("ffd1c000.i2c")
if alias is not None:
    globals()[alias + "_SCL"] = GPIOA_15
    globals()[alias + "_SDA"] = GPIOA_14
    i2cPorts.append((int(alias[3]), GPIOA_15, GPIOA_14))

i2cPorts = tuple(i2cPorts)



"""
`busio` - Bus protocol support like I2C and SPI
=================================================

See `CircuitPython:busio` in CircuitPython for more details.

* Author(s): cefn
"""

try:
    import threading
except ImportError:
    threading = None

import adafruit_platformdetect.constants.boards as ap_board
import adafruit_platformdetect.constants.chips as ap_chip
from adafruit_blinka import Enum, Lockable, agnostic
from adafruit_blinka.agnostic import board_id, detector

# pylint: disable=import-outside-toplevel,too-many-branches,too-many-statements
# pylint: disable=too-many-arguments,too-many-function-args


class I2C(Lockable):
    """
    Busio I2C Class for CircuitPython Compatibility. Used
    for both MicroPython and Linux.
    """

    def __init__(self, scl, sda, frequency=100000):
        self.init(scl, sda, frequency)

    def init(self, scl, sda, frequency):
        """Initialization"""
        self.deinit()
        if detector.board.ftdi_ft232h:
            from adafruit_blinka.microcontroller.ft232h.i2c import I2C as _I2C

            self._i2c = _I2C(frequency=frequency)
            return
        if detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.i2c import I2C as _I2C

            self._i2c = _I2C(frequency=frequency)
            return
        if detector.board.microchip_mcp2221:
            from adafruit_blinka.microcontroller.mcp2221.i2c import I2C as _I2C

            self._i2c = _I2C(frequency=frequency)
            return
        if detector.board.greatfet_one:
            from adafruit_blinka.microcontroller.nxp_lpc4330.i2c import I2C as _I2C

            self._i2c = _I2C(frequency=frequency)
            return
        if detector.board.any_embedded_linux:
            from adafruit_blinka.microcontroller.generic_linux.i2c import I2C as _I2C
        else:
            from machine import I2C as _I2C
        from microcontroller.pin import i2cPorts

        for portId, portScl, portSda in i2cPorts:
            try:
                if scl == portScl and sda == portSda:
                    self._i2c = _I2C(portId, mode=_I2C.MASTER, baudrate=frequency)
                    break
            except RuntimeError:
                pass
        else:
            raise ValueError(
                "No Hardware I2C on (scl,sda)={}\nValid I2C ports: {}".format(
                    (scl, sda), i2cPorts
                )
            )
        if threading is not None:
            self._lock = threading.RLock()

    def deinit(self):
        """Deinitialization"""
        try:
            del self._i2c
        except AttributeError:
            pass

    def __enter__(self):
        if threading is not None:
            self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if threading is not None:
            self._lock.release()
        self.deinit()

    def scan(self):
        """Scan for attached devices"""
        return self._i2c.scan()

    def readfrom_into(self, address, buffer, *, start=0, end=None):
        """Read from a device at specified address into a buffer"""
        if start != 0 or end is not None:
            if end is None:
                end = len(buffer)
            buffer = memoryview(buffer)[start:end]
        stop = True  # remove for efficiency later
        return self._i2c.readfrom_into(address, buffer, stop=stop)

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        """Write to a device at specified address from a buffer"""
        if isinstance(buffer, str):
            buffer = bytes([ord(x) for x in buffer])
        if start != 0 or end is not None:
            if end is None:
                return self._i2c.writeto(address, memoryview(buffer)[start:], stop=stop)
            return self._i2c.writeto(address, memoryview(buffer)[start:end], stop=stop)
        return self._i2c.writeto(address, buffer, stop=stop)

    def writeto_then_readfrom(
        self,
        address,
        buffer_out,
        buffer_in,
        *,
        out_start=0,
        out_end=None,
        in_start=0,
        in_end=None,
        stop=False
    ):
        """ "Write to a device at specified address from a buffer then read
        from a device at specified address into a buffer
        """
        return self._i2c.writeto_then_readfrom(
            address,
            buffer_out,
            buffer_in,
            out_start=out_start,
            out_end=out_end,
            in_start=in_start,
            in_end=in_end,
            stop=stop,
        )


class SPI(Lockable):
    """
    Busio SPI Class for CircuitPython Compatibility. Used
    for both MicroPython and Linux.
    """

    def __init__(self, clock, MOSI=None, MISO=None):
        self.deinit()
        if detector.board.ftdi_ft232h:
            from adafruit_blinka.microcontroller.ft232h.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.ft232h.pin import SCK, MOSI, MISO

            self._spi = _SPI()
            self._pins = (SCK, MOSI, MISO)
            return
        if detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.nova.pin import SCK, MOSI, MISO

            self._spi = _SPI(clock)
            self._pins = (SCK, MOSI, MISO)
            return
        if detector.board.greatfet_one:
            from adafruit_blinka.microcontroller.nxp_lpc4330.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.nxp_lpc4330.pin import SCK, MOSI, MISO

            self._spi = _SPI()
            self._pins = (SCK, MOSI, MISO)
            return
        if detector.board.any_embedded_linux:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        else:
            from machine import SPI as _SPI
        from microcontroller.pin import spiPorts

        for portId, portSck, portMosi, portMiso in spiPorts:
            if (
                (clock == portSck)
                and MOSI in (portMosi, None)  # Clock is required!
                and MISO in (portMiso, None)  # But can do with just output
            ):  # Or just input
                self._spi = _SPI(portId)
                self._pins = (portSck, portMosi, portMiso)
                break
        else:
            raise ValueError(
                "No Hardware SPI on (SCLK, MOSI, MISO)={}\nValid SPI ports:{}".format(
                    (clock, MOSI, MISO), spiPorts
                )
            )

    def configure(self, baudrate=100000, polarity=0, phase=0, bits=8):
        """Update the configuration"""
        if detector.board.any_raspberry_pi or detector.board.any_raspberry_pi_40_pin:
            from adafruit_blinka.microcontroller.bcm283x.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif detector.board.BEAGLEBONE_AI:
            from adafruit_blinka.microcontroller.dra74x.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif detector.board.any_beaglebone:
            from adafruit_blinka.microcontroller.am335x.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif detector.board.any_orange_pi and detector.chip.id == ap_chip.SUN8I:
            from adafruit_blinka.microcontroller.allwinner.h3.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif detector.board.any_nanopi and detector.chip.id == ap_chip.SUN8I:
            from adafruit_blinka.microcontroller.allwinner.h3.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.GIANT_BOARD:
            from adafruit_blinka.microcontroller.sama5.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.CORAL_EDGE_TPU_DEV:
            from adafruit_blinka.microcontroller.nxp_imx8m.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.CORAL_EDGE_TPU_DEV_MINI:
            from adafruit_blinka.microcontroller.mt8167.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.ODROID_C2:
            from adafruit_blinka.microcontroller.amlogic.s905.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.ODROID_C4:
            from adafruit_blinka.microcontroller.amlogic.s905x3.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.ODROID_XU4:
            from adafruit_blinka.microcontroller.samsung.exynos5422.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.DRAGONBOARD_410C:
            from adafruit_blinka.microcontroller.snapdragon.apq8016.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.JETSON_NANO:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t210.pin import Pin
        elif board_id == ap_board.JETSON_TX1:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t210.pin import Pin
        elif board_id == ap_board.JETSON_TX2:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t186.pin import Pin
        elif board_id == ap_board.JETSON_XAVIER:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t194.pin import Pin
        elif board_id == ap_board.JETSON_NX:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t194.pin import Pin
        elif detector.board.ROCK_PI_S:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.rockchip.rk3308.pin import Pin
        elif detector.board.SIFIVE_UNLEASHED:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.hfu540.pin import Pin
        elif detector.board.ftdi_ft232h:
            from adafruit_blinka.microcontroller.ft232h.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.ft232h.pin import Pin
        elif detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.nova.pin import Pin
        elif detector.board.greatfet_one:
            from adafruit_blinka.microcontroller.nxp_lpc4330.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.nxp_lpc4330.pin import Pin
        elif board_id in (
            ap_board.PINE64,
            ap_board.PINEBOOK,
            ap_board.PINEPHONE,
            ap_board.SOPINE,
        ):
            from adafruit_blinka.microcontroller.allwinner.a64.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.CLOCKWORK_CPI3:
            from adafruit_blinka.microcontroller.allwinner.a33.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.ONION_OMEGA2:
            from adafruit_blinka.microcontroller.mips24kec.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif detector.board.any_lubancat and detector.chip.id == ap_chip.IMX6ULL:
            from adafruit_blinka.microcontroller.nxp_imx6ull.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        else:
            from machine import SPI as _SPI
            from machine import Pin

        if self._locked:
            # TODO check if #init ignores MOSI=None rather than unsetting, to save _pinIds attribute
            self._spi.init(
                baudrate=baudrate,
                polarity=polarity,
                phase=phase,
                bits=bits,
                firstbit=_SPI.MSB,
                sck=Pin(self._pins[0].id),
                mosi=Pin(self._pins[1].id),
                miso=Pin(self._pins[2].id),
            )
        else:
            raise RuntimeError("First call try_lock()")

    def deinit(self):
        """Deinitialization"""
        self._spi = None
        self._pinIds = None

    @property
    def frequency(self):
        """Return the baud rate if implemented"""
        try:
            return self._spi.frequency
        except AttributeError:
            raise NotImplementedError(
                "Frequency attribute not implemented for this platform"
            ) from AttributeError

    def write(self, buf, start=0, end=None):
        """Write to the SPI device"""
        return self._spi.write(buf, start, end)

    def readinto(self, buf, start=0, end=None, write_value=0):
        """Read from the SPI device into a buffer"""
        return self._spi.readinto(buf, start, end, write_value=write_value)

    def write_readinto(
        self, buffer_out, buffer_in, out_start=0, out_end=None, in_start=0, in_end=None
    ):
        """Write to the SPI device and read from the SPI device into a buffer"""
        return self._spi.write_readinto(
            buffer_out, buffer_in, out_start, out_end, in_start, in_end
        )


class UART(Lockable):
    """
    Busio UART Class for CircuitPython Compatibility. Used
    for MicroPython and a few other non-Linux boards.
    """

    class Parity(Enum):
        """Parity Enumeration"""

        pass  # pylint: disable=unnecessary-pass

    Parity.ODD = Parity()
    Parity.EVEN = Parity()

    def __init__(
        self,
        tx,
        rx,
        baudrate=9600,
        bits=8,
        parity=None,
        stop=1,
        timeout=1000,
        receiver_buffer_size=64,
        flow=None,
    ):
        if detector.board.any_embedded_linux:
            raise RuntimeError(
                "busio.UART not supported on this platform. Please use pyserial instead."
            )
        if detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.uart import UART as _UART
        elif detector.board.greatfet_one:
            from adafruit_blinka.microcontroller.nxp_lpc4330.uart import UART as _UART
        else:
            from machine import UART as _UART

        if detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.pin import uartPorts
        else:
            from microcontroller.pin import uartPorts

        self.baudrate = baudrate

        if flow is not None:  # default 0
            raise NotImplementedError(
                "Parameter '{}' unsupported on {}".format("flow", agnostic.board_id)
            )

        # translate parity flag for Micropython
        if parity is UART.Parity.ODD:
            parity = 1
        elif parity is UART.Parity.EVEN:
            parity = 0
        elif parity is None:
            pass
        else:
            raise ValueError("Invalid parity")

        # check tx and rx have hardware support
        for portId, portTx, portRx in uartPorts:  #
            if portTx == tx and portRx == rx:
                self._uart = _UART(
                    portId,
                    baudrate,
                    bits=bits,
                    parity=parity,
                    stop=stop,
                    timeout=timeout,
                    read_buf_len=receiver_buffer_size,
                )
                break
        else:
            raise ValueError(
                "No Hardware UART on (tx,rx)={}\nValid UART ports: {}".format(
                    (tx, rx), uartPorts
                )
            )

    def deinit(self):
        """Deinitialization"""
        if detector.board.binho_nova:
            self._uart.deinit()
        self._uart = None

    def read(self, nbytes=None):
        """Read from the UART"""
        return self._uart.read(nbytes)

    def readinto(self, buf, nbytes=None):
        """Read from the UART into a buffer"""
        return self._uart.readinto(buf, nbytes)

    def readline(self):
        """Read a line of characters up to a newline charater from the UART"""
        return self._uart.readline()

    def write(self, buf):
        """Write to the UART from a buffer"""
        return self._uart.write(buf)


class OneWire:
    """
    Stub class for OneWire, which is currently not implemented
    """

    def __init__(self, pin):
        raise NotImplementedError("OneWire has not been implemented")

    def deinit(self):
        """
        Deinitialize the OneWire bus and release any hardware resources for reuse.
        """
        raise NotImplementedError("OneWire has not been implemented")

    def reset(self):
        """
        Reset the OneWire bus and read presence
        """
        raise NotImplementedError("OneWire has not been implemented")

    def read_bit(self):
        """
        Read in a bit
        """
        raise NotImplementedError("OneWire has not been implemented")

    def write_bit(self, value):
        """
        Write out a bit based on value.
        """
        raise NotImplementedError("OneWire has not been implemented")



"""Pin definitions for the Pine64."""

from adafruit_blinka.microcontroller.allwinner.a64 import pin

D2 = pin.PH3
D3 = pin.PH2
D4 = pin.PL10
D5 = pin.PH5
D6 = pin.PH6
D7 = pin.PH7
D8 = pin.PC3
D9 = pin.PC1
D10 = pin.PC0
D11 = pin.PC2
D12 = pin.PC4
D13 = pin.PC5
D14 = pin.PB0
D15 = pin.PB1
D16 = pin.PC6
D17 = pin.PC7
D18 = pin.PC8
D19 = pin.PC9
D20 = pin.PC10
D21 = pin.PC11
D22 = pin.PC12
D23 = pin.PC13
D24 = pin.PC14
D25 = pin.PC15
D26 = pin.PC16
D27 = pin.PH9

SDA = D2
SCL = D3

SCL2 = pin.PL8
SDA2 = pin.PL9

SCLK = D11
MOSI = D10
MISO = D9
CS = D8
SCK = SCLK

UART_TX = D14
UART_RX = D15

UART3_TX = pin.PD0
UART3_RX = pin.PD1

UART4_TX = pin.PD2
UART4_RX = pin.PD3



"""Pin definitions for the Odroid C2."""

from adafruit_blinka.microcontroller.amlogic.s905 import pin

GPIOX_0 = pin.GPIO228
GPIOX_1 = pin.GPIO229
GPIOX_2 = pin.GPIO230
GPIOX_3 = pin.GPIO231
GPIOX_4 = pin.GPIO232
GPIOX_5 = pin.GPIO233
GPIOX_6 = pin.GPIO234
GPIOX_7 = pin.GPIO235
GPIOX_8 = pin.GPIO236
GPIOX_9 = pin.GPIO237
GPIOX_10 = pin.GPIO238
GPIOX_11 = pin.GPIO239
GPIOX_12 = pin.GPIO240
GPIOX_13 = pin.GPIO241
GPIOX_19 = pin.GPIO247
GPIOX_21 = pin.GPIO249

GPIODV_24 = pin.GPIO205
GPIODV_25 = pin.GPIO206
GPIODV_26 = pin.GPIO207
GPIODV_27 = pin.GPIO208
GPIODV_28 = pin.GPIO209
GPIODV_29 = pin.GPIO210

GPIOY_0 = pin.GPIO211
GPIOY_1 = pin.GPIO212
GPIOY_3 = pin.GPIO214
GPIOY_7 = pin.GPIO218
GPIOY_8 = pin.GPIO219
GPIOY_12 = pin.GPIO223
GPIOY_13 = pin.GPIO224
GPIOY_14 = pin.GPIO225

GPIOA0_0 = pin.GPIO122
GPIOA0_1 = pin.GPIO123
GPIOA0_2 = pin.GPIO124
GPIOA0_3 = pin.GPIO125
GPIOA0_4 = pin.GPIO126
GPIOA0_5 = pin.GPIO127
GPIOA0_6 = pin.GPIO128
GPIOA0_7 = pin.GPIO129
GPIOA0_8 = pin.GPIO130
GPIOA0_9 = pin.GPIO131
GPIOA0_10 = pin.GPIO132
GPIOA0_11 = pin.GPIO133
GPIOA0_12 = pin.GPIO134
GPIOA0_13 = pin.GPIO135

SDA = pin.I2C0_SDA
SCL = pin.I2C0_SCL

SCLK = pin.SPI0_SCLK
MOSI = pin.SPI0_MOSI
MISO = pin.SPI0_MISO
SPI_CS0 = pin.GPIO229
SPI_CS1 = pin.GPIO225

D0 = GPIOX_19
D1 = GPIOX_10
D2 = GPIOX_11
D3 = GPIOX_9
D4 = GPIOX_8
D5 = GPIOX_5
D6 = GPIOX_3
D7 = GPIOX_21
D10 = GPIOX_1
D11 = GPIOY_14
D12 = GPIOX_7
D13 = GPIOX_4
D14 = GPIOX_2
D21 = GPIOX_0
D22 = GPIOY_8
D23 = GPIOX_6
D24 = GPIOY_3
D26 = GPIOY_13
D27 = GPIOY_7



"""SPI Class for FT232H"""
from adafruit_blinka.microcontroller.ft232h.pin import Pin

# pylint: disable=protected-access
class SPI:
    """Custom SPI Class for FT232H"""

    MSB = 0

    def __init__(self):
        # pylint: disable=import-outside-toplevel
        from pyftdi.spi import SpiController

        # pylint: enable=import-outside-toplevel

        self._spi = SpiController(cs_count=1)
        self._spi.configure("ftdi://ftdi:ft232h/1")
        self._port = self._spi.get_port(0)
        self._port.set_frequency(100000)
        self._port._cpol = 0
        self._port._cpha = 0
        # Change GPIO controller to SPI
        Pin.ft232h_gpio = self._spi.get_gpio()

    # pylint: disable=too-many-arguments,unused-argument
    def init(
        self,
        baudrate=100000,
        polarity=0,
        phase=0,
        bits=8,
        firstbit=MSB,
        sck=None,
        mosi=None,
        miso=None,
    ):
        """Initialize the Port"""
        self._port.set_frequency(baudrate)
        # FTDI device can only support mode 0 and mode 2
        # due to the limitation of MPSSE engine.
        # This means CPHA must = 0
        self._port._cpol = polarity
        if phase != 0:
            raise ValueError("Only SPI phase 0 is supported by FT232H.")
        self._port._cpha = phase

    # pylint: enable=too-many-arguments

    @property
    def frequency(self):
        """Return the current frequency"""
        return self._port.frequency

    def write(self, buf, start=0, end=None):
        """Write data from the buffer to SPI"""
        end = end if end else len(buf)
        chunks, rest = divmod(end - start, self._spi.PAYLOAD_MAX_LENGTH)
        for i in range(chunks):
            chunk_start = start + i * self._spi.PAYLOAD_MAX_LENGTH
            chunk_end = chunk_start + self._spi.PAYLOAD_MAX_LENGTH
            self._port.write(buf[chunk_start:chunk_end])
        if rest:
            rest_start = start + chunks * self._spi.PAYLOAD_MAX_LENGTH
            self._port.write(buf[rest_start:end])

    # pylint: disable=unused-argument
    def readinto(self, buf, start=0, end=None, write_value=0):
        """Read data from SPI and into the buffer"""
        end = end if end else len(buf)
        buffer_out = [write_value] * (end - start)
        result = self._port.exchange(buffer_out, end - start, duplex=True)
        for i, b in enumerate(result):
            buf[start + i] = b

    # pylint: enable=unused-argument

    # pylint: disable=too-many-arguments
    def write_readinto(
        self, buffer_out, buffer_in, out_start=0, out_end=None, in_start=0, in_end=None
    ):
        """Perform a half-duplex write from buffer_out and then
        read data into buffer_in
        """
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)
        result = self._port.exchange(
            buffer_out[out_start:out_end], in_end - in_start, duplex=True
        )
        for i, b in enumerate(result):
            buffer_in[in_start + i] = b

    # pylint: enable=too-many-arguments



"""A Pin class for use with Rockchip RK3308."""

from adafruit_blinka.microcontroller.generic_linux.sysfs_pin import Pin

GPIO0_A0 = Pin(0)
GPIO0_A1 = Pin(1)
GPIO0_A2 = Pin(2)
GPIO0_A3 = Pin(3)
GPIO0_A4 = Pin(4)
GPIO0_A5 = Pin(5)
GPIO0_A6 = Pin(6)
GPIO0_A7 = Pin(7)
GPIO0_B0 = Pin(8)
GPIO0_B1 = Pin(9)
GPIO0_B2 = Pin(10)
GPIO0_B3 = Pin(11)
GPIO0_B4 = Pin(12)
GPIO0_B5 = Pin(13)
GPIO0_B6 = Pin(14)
GPIO0_B7 = Pin(15)
GPIO0_C0 = Pin(16)
GPIO0_C1 = Pin(17)
GPIO0_C2 = Pin(18)
GPIO0_C3 = Pin(19)
GPIO0_C4 = Pin(20)
GPIO0_C5 = Pin(21)
GPIO0_C6 = Pin(22)
GPIO0_C7 = Pin(23)
GPIO0_D0 = Pin(24)
GPIO0_D1 = Pin(25)
GPIO0_D2 = Pin(26)
GPIO0_D3 = Pin(27)
GPIO0_D4 = Pin(28)
GPIO0_D5 = Pin(29)
GPIO0_D6 = Pin(30)
GPIO0_D7 = Pin(31)
GPIO1_A0 = Pin(32)
GPIO1_A1 = Pin(33)
GPIO1_A2 = Pin(34)
GPIO1_A3 = Pin(35)
GPIO1_A4 = Pin(36)
GPIO1_A5 = Pin(37)
GPIO1_A6 = Pin(38)
GPIO1_A7 = Pin(39)
GPIO1_B0 = Pin(40)
GPIO1_B1 = Pin(41)
GPIO1_B2 = Pin(42)
GPIO1_B3 = Pin(43)
GPIO1_B4 = Pin(44)
GPIO1_B5 = Pin(45)
GPIO1_B6 = Pin(46)
GPIO1_B7 = Pin(47)
GPIO1_C0 = Pin(48)
GPIO1_C1 = Pin(49)
GPIO1_C2 = Pin(50)
GPIO1_C3 = Pin(51)
GPIO1_C4 = Pin(52)
GPIO1_C5 = Pin(53)
GPIO1_C6 = Pin(54)
GPIO1_C7 = Pin(55)
GPIO1_D0 = Pin(56)
GPIO1_D1 = Pin(57)
GPIO1_D2 = Pin(58)
GPIO1_D3 = Pin(59)
GPIO1_D4 = Pin(60)
GPIO1_D5 = Pin(61)
GPIO1_D6 = Pin(62)
GPIO1_D7 = Pin(63)
GPIO2_A0 = Pin(64)
GPIO2_A1 = Pin(65)
GPIO2_A2 = Pin(66)
GPIO2_A3 = Pin(67)
GPIO2_A4 = Pin(68)
GPIO2_A5 = Pin(69)
GPIO2_A6 = Pin(70)
GPIO2_A7 = Pin(71)
GPIO2_B0 = Pin(72)
GPIO2_B1 = Pin(73)
GPIO2_B2 = Pin(74)
GPIO2_B3 = Pin(75)
GPIO2_B4 = Pin(76)
GPIO2_B5 = Pin(77)
GPIO2_B6 = Pin(78)
GPIO2_B7 = Pin(79)
GPIO2_C0 = Pin(80)
GPIO2_C1 = Pin(81)
GPIO2_C2 = Pin(82)
GPIO2_C3 = Pin(83)
GPIO2_C4 = Pin(84)
GPIO2_C5 = Pin(85)
GPIO2_C6 = Pin(86)
GPIO2_C7 = Pin(87)
GPIO2_D0 = Pin(88)
GPIO2_D1 = Pin(89)
GPIO2_D2 = Pin(90)
GPIO2_D3 = Pin(91)
GPIO2_D4 = Pin(92)
GPIO2_D5 = Pin(93)
GPIO2_D6 = Pin(94)
GPIO2_D7 = Pin(95)
GPIO3_A0 = Pin(96)
GPIO3_A1 = Pin(97)
GPIO3_A2 = Pin(98)
GPIO3_A3 = Pin(99)
GPIO3_A4 = Pin(100)
GPIO3_A5 = Pin(101)
GPIO3_A6 = Pin(102)
GPIO3_A7 = Pin(103)
GPIO3_B0 = Pin(104)
GPIO3_B1 = Pin(105)
GPIO3_B2 = Pin(106)
GPIO3_B3 = Pin(107)
GPIO3_B4 = Pin(108)
GPIO3_B5 = Pin(109)
GPIO3_B6 = Pin(110)
GPIO3_B7 = Pin(111)
GPIO3_C0 = Pin(112)
GPIO3_C1 = Pin(113)
GPIO3_C2 = Pin(114)
GPIO3_C3 = Pin(115)
GPIO3_C4 = Pin(116)
GPIO3_C5 = Pin(117)
GPIO3_C6 = Pin(118)
GPIO3_C7 = Pin(119)
GPIO3_D0 = Pin(120)
GPIO3_D1 = Pin(121)
GPIO3_D2 = Pin(122)
GPIO3_D3 = Pin(123)
GPIO3_D4 = Pin(124)
GPIO3_D5 = Pin(125)
GPIO3_D6 = Pin(126)
GPIO3_D7 = Pin(127)
ADC_IN0 = 1


# I2C
I2C0_SDA = GPIO1_D0
I2C0_SCL = GPIO1_D1
I2C1_SDA = GPIO0_B3
I2C1_SCL = GPIO0_B4
I2C2_SDA = GPIO2_A2
I2C2_SCL = GPIO2_A3
I2C3_SDA = GPIO0_B7
I2C3_SCL = GPIO0_C0

# SPI
SPI2_CS = GPIO1_D1
SPI2_SCLK = GPIO1_D0
SPI2_MISO = GPIO1_C6
SPI2_MOSI = GPIO1_C7

# UART
UART0_TX = GPIO2_A1
UART0_RX = GPIO2_A0
UART1_TX = GPIO1_D1
UART1_RX = GPIO1_D0
UART2_TX = GPIO1_C7
UART2_RX = GPIO1_C6

# PWM
PWM2 = GPIO0_B7
PWM3 = GPIO0_C0

# ordered as i2cId, SCL, SDA
i2cPorts = (
    (0, I2C0_SCL, I2C0_SDA),
    (1, I2C1_SCL, I2C1_SDA),
    (2, I2C2_SCL, I2C2_SDA),
    (3, I2C3_SCL, I2C3_SDA),
)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((2, SPI2_SCLK, SPI2_MOSI, SPI2_MISO),)

# SysFS pwm outputs, pwm channel and pin in first tuple
pwmOuts = (
    ((1, 0), PWM2),
    ((2, 0), PWM3),
)

# SysFS analog inputs, Ordered as analog analogInId, device, and channel
analogIns = ((ADC_IN0, 0, 0),)



"""I2C Class for FT232H"""
from adafruit_blinka.microcontroller.ft232h.pin import Pin


class I2C:
    """Custom I2C Class for FT232H"""

    def __init__(self, *, frequency=400000):
        # change GPIO controller to I2C
        # pylint: disable=import-outside-toplevel
        from pyftdi.i2c import I2cController

        # pylint: enable=import-outside-toplevel

        self._i2c = I2cController()
        self._i2c.configure("ftdi://ftdi:ft232h/1", frequency=frequency)
        Pin.ft232h_gpio = self._i2c.get_gpio()

    def scan(self):
        """Perform an I2C Device Scan"""
        return [addr for addr in range(0x79) if self._i2c.poll(addr)]

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        """Write data from the buffer to an address"""
        end = end if end else len(buffer)
        port = self._i2c.get_port(address)
        port.write(buffer[start:end], relax=stop)

    def readfrom_into(self, address, buffer, *, start=0, end=None, stop=True):
        """Read data from an address and into the buffer"""
        end = end if end else len(buffer)
        port = self._i2c.get_port(address)
        result = port.read(len(buffer[start:end]), relax=stop)
        for i, b in enumerate(result):
            buffer[start + i] = b

    # pylint: disable=unused-argument
    def writeto_then_readfrom(
        self,
        address,
        buffer_out,
        buffer_in,
        *,
        out_start=0,
        out_end=None,
        in_start=0,
        in_end=None,
        stop=False
    ):
        """Write data from buffer_out to an address and then
        read data from an address and into buffer_in
        """
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)
        port = self._i2c.get_port(address)
        result = port.exchange(
            buffer_out[out_start:out_end], in_end - in_start, relax=True
        )
        for i, b in enumerate(result):
            buffer_in[in_start + i] = b

    # pylint: enable=unused-argument



"""
`analogio` - Analog input and output control
=================================================
See `CircuitPython:analogio` in CircuitPython for more details.
* Author(s): Carter Nelson
"""

from adafruit_blinka.microcontroller.mcp2221.pin import Pin
from adafruit_blinka import ContextManaged


class AnalogIn(ContextManaged):
    """Analog Input Class"""

    def __init__(self, pin):
        self._pin = Pin(pin.id)
        self._pin.init(mode=Pin.ADC)

    @property
    def value(self):
        """Read the ADC and return the value"""
        return self._pin.value()

    # pylint: disable=no-self-use
    @value.setter
    def value(self, value):
        # emulate what CircuitPython does
        raise AttributeError("'AnalogIn' object has no attribute 'value'")

    # pylint: enable=no-self-use

    def deinit(self):
        del self._pin


class AnalogOut(ContextManaged):
    """Analog Output Class"""

    def __init__(self, pin):
        self._pin = Pin(pin.id)
        self._pin.init(mode=Pin.DAC)

    @property
    def value(self):
        """Return an error. This is output only."""
        # emulate what CircuitPython does
        raise AttributeError("unreadable attribute")

    @value.setter
    def value(self, value):
        self._pin.value(value)

    def deinit(self):
        del self._pin



"""Based on https://raw.githubusercontent.com/micropython/micropython-lib/cfa1b9cce0c93a3115bbff3886c9bbcddd9e8047/unittest/unittest.py """
import sys


class SkipTest(Exception):
    pass


raiseException = False
raiseBaseException = True


class AssertRaisesContext:
    def __init__(self, exc):
        self.expected = exc

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            assert False, "%r not raised" % self.expected
        if issubclass(exc_type, self.expected):
            return True
        return False


class TestCase:
    def fail(self, msg=""):
        assert False, msg

    def assertEqual(self, x, y, msg=""):
        if not msg:
            msg = "%r vs (expected) %r" % (x, y)
        assert x == y, msg

    def assertNotEqual(self, x, y, msg=""):
        if not msg:
            msg = "%r not expected to be equal %r" % (x, y)
        assert x != y, msg

    def assertAlmostEqual(self, x, y, places=None, msg="", delta=None):
        if x == y:
            return
        if delta is not None and places is not None:
            raise TypeError("specify delta or places not both")

        if delta is not None:
            if abs(x - y) <= delta:
                return
            if not msg:
                msg = "%r != %r within %r delta" % (x, y, delta)
        else:
            if places is None:
                places = 7
            if round(abs(y - x), places) == 0:
                return
            if not msg:
                msg = "%r != %r within %r places" % (x, y, places)

        assert False, msg

    def assertNotAlmostEqual(self, x, y, places=None, msg="", delta=None):
        if delta is not None and places is not None:
            raise TypeError("specify delta or places not both")

        if delta is not None:
            if not (x == y) and abs(x - y) > delta:
                return
            if not msg:
                msg = "%r == %r within %r delta" % (x, y, delta)
        else:
            if places is None:
                places = 7
            if not (x == y) and round(abs(y - x), places) != 0:
                return
            if not msg:
                msg = "%r == %r within %r places" % (x, y, places)

        assert False, msg

    def assertIs(self, x, y, msg=""):
        if not msg:
            msg = "%r is not %r" % (x, y)
        assert x is y, msg

    def assertIsNot(self, x, y, msg=""):
        if not msg:
            msg = "%r is %r" % (x, y)
        assert x is not y, msg

    def assertIsNone(self, x, msg=""):
        if not msg:
            msg = "%r is not None" % x
        assert x is None, msg

    def assertIsNotNone(self, x, msg=""):
        if not msg:
            msg = "%r is None" % x
        assert x is not None, msg

    def assertTrue(self, x, msg=""):
        if not msg:
            msg = "Expected %r to be True" % x
        assert x, msg

    def assertFalse(self, x, msg=""):
        if not msg:
            msg = "Expected %r to be False" % x
        assert not x, msg

    def assertIn(self, x, y, msg=""):
        if not msg:
            msg = "Expected %r to be in %r" % (x, y)
        assert x in y, msg

    def assertIsInstance(self, x, y, msg=""):
        assert isinstance(x, y), msg

    def assertRaises(self, exc, func=None, *args, **kwargs):
        if func is None:
            return AssertRaisesContext(exc)

        try:
            func(*args, **kwargs)
            assert False, "%r not raised" % exc
        except Exception as e:
            if isinstance(e, exc):
                return
            raise


def skip(msg):
    def _decor(fun):
        # We just replace original fun with _inner
        def _inner(self):
            raise SkipTest(msg)

        return _inner

    return _decor


def skipIf(cond, msg):
    if not cond:
        return lambda x: x
    return skip(msg)


def skipUnless(cond, msg):
    if cond:
        return lambda x: x
    return skip(msg)


class TestSuite:
    def __init__(self):
        self.tests = []

    def addTest(self, cls):
        self.tests.append(cls)


class TestRunner:
    def run(self, suite):
        res = TestResult()
        for c in suite.tests:
            run_class(c, res)

        print("Ran %d tests\n" % res.testsRun)
        if res.failuresNum > 0 or res.errorsNum > 0:
            print("FAILED (failures=%d, errors=%d)" % (res.failuresNum, res.errorsNum))
        else:
            msg = "OK"
            if res.skippedNum > 0:
                msg += " (%d skipped)" % res.skippedNum
            print(msg)

        return res


class TestResult:
    def __init__(self):
        self.errorsNum = 0
        self.failuresNum = 0
        self.skippedNum = 0
        self.testsRun = 0

    def wasSuccessful(self):
        return self.errorsNum == 0 and self.failuresNum == 0


# TODO: Uncompliant
def run_class(c, test_result):
    o = c()
    set_up = getattr(o, "setUp", lambda: None)
    tear_down = getattr(o, "tearDown", lambda: None)
    for name in dir(o):
        if name.startswith("test"):
            print("%s (%s) ..." % (name, c.__qualname__), end="")
            m = getattr(o, name)
            set_up()
            try:
                test_result.testsRun += 1
                m()
                print(" ok")
            except SkipTest as e:
                print(" skipped:", e.args[0])
                test_result.skippedNum += 1
            except Exception as e:  # user exception
                print(" FAIL")
                if raiseException:
                    raise
                else:
                    print(e)
                    test_result.failuresNum += 1
                    continue
            except BaseException as e:  # system exception
                print(" FAIL")
                if raiseBaseException:
                    raise
                else:
                    print(e)
                    test_result.failuresNum += 1
                    continue
            finally:
                tear_down()


def main(module="__main__"):
    def test_cases(m):
        for tn in dir(m):
            c = getattr(m, tn)
            if (
                isinstance(c, object)
                and isinstance(c, type)
                and issubclass(c, TestCase)
            ):
                yield c

    m = __import__(module)  # changed to permit non-top-level testing modules
    suite = TestSuite()
    for c in test_cases(m):
        suite.addTest(c)
    runner = TestRunner()
    result = runner.run(suite)



"""
`analogio` - Analog input control
=================================================
Read the SysFS ADC using IIO (Industrial Input/Output) and return the value

See `CircuitPython:analogio` in CircuitPython for more details.
* Author(s): Melissa LeBlanc-Williams
"""

import os
from adafruit_blinka import ContextManaged

try:
    from microcontroller.pin import analogIns
except ImportError:
    raise RuntimeError("No Analog Inputs defined for this board") from ImportError


class AnalogIn(ContextManaged):
    """Analog Input Class"""

    # Sysfs paths
    _sysfs_path = "/sys/bus/iio/devices/"
    _device_path = "iio:device{}"

    # Channel paths
    _channel_path = "in_voltage{}_raw"
    _scale_path = "in_voltage_scale"

    def __init__(self, adc_id):
        """Instantiate an AnalogIn object and verify the sysfs IIO
        corresponding to the specified channel and pin.

        Args:
            adc_id (int): Analog Input ID as defined in microcontroller.pin

        Returns:
            AnalogIn: AnalogIn object.

        Raises:
            TypeError: if `channel` or `pin` types are invalid.
            ValueError: if AnalogIn channel does not exist.

        """

        self.id = adc_id
        self._device = None
        self._channel = None
        self._open(adc_id)

    def __enter__(self):
        return self

    def _open(self, adc_id):
        self._device = None
        for adcpair in analogIns:
            if adcpair[0] == adc_id:
                self._device = adcpair[1]
                self._channel = adcpair[2]

        if self._device is None:
            raise RuntimeError("No AnalogIn device found for the given ID")

        device_path = os.path.join(
            self._sysfs_path, self._device_path.format(self._device)
        )

        if not os.path.isdir(device_path):
            raise ValueError(
                "AnalogIn device does not exist, check that the required modules are loaded."
            )

    @property
    def value(self):
        """Read the ADC and return the value as an integer"""
        path = os.path.join(
            self._sysfs_path,
            self._device_path.format(self._device),
            self._channel_path.format(self._channel),
        )

        with open(path, "r") as analog_in:
            return int(analog_in.read().strip())

    # pylint: disable=no-self-use
    @value.setter
    def value(self, value):
        # emulate what CircuitPython does
        raise AttributeError("'AnalogIn' object has no attribute 'value'")

    # pylint: enable=no-self-use

    def deinit(self):
        self._device = None
        self._channel = None



"""Pin definitions for Jetson TX2."""

from adafruit_blinka.microcontroller.tegra.t186 import pin

SDA = pin.SDA
SCL = pin.SCL
SDA_1 = pin.SDA_1
SCL_1 = pin.SCL_1

D4 = pin.J04
D5 = pin.J06
D6 = pin.AA02
D7 = None
D8 = pin.N06
D9 = pin.N04
D10 = pin.N05
D11 = pin.N03
D12 = pin.AA01
D13 = pin.I05
D16 = pin.T03
D17 = pin.T02
D18 = pin.J00
D19 = pin.J03
D20 = pin.J02
D21 = pin.J01
D22 = pin.P17
D23 = pin.AA00
D24 = pin.Y01
D25 = pin.P16
D26 = pin.I04
D27 = pin.J05

CE1 = None
CE0 = D8
MISO = D9
MOSI = D10
SCLK = D11
SCK = D11



"""Pin definitions for the LubanCat IMX6ULL."""

from adafruit_blinka.microcontroller.nxp_imx6ull import pin

# Pro board pin header CN4 named GPIO_PAx, pin header CN5 named GPIO_PBx
# Mini board pin header CN3 named GPIO_PCx, pin header CN4 named GPIO_PDx

# Board pin name [= alias] = RPI name [= alias] = pin name
GPIO_PC3 = I2C2_SDA = D2 = SDA = pin.I2C2_SDA
GPIO_PC5 = I2C2_SCL = D3 = SCL = pin.I2C2_SCL

GPIO_PC7 = D4 = pin.GPIO27
GPIO_PC8 = D14 = TXD = pin.UART3_TXD
GPIO_PC10 = D15 = RXD = pin.UART3_RXD
GPIO_PC11 = ADC_IN3 = A3 = D17 = pin.GPIO3
GPIO_PC12 = D18 = pin.GPIO112
GPIO_PC13 = ADC_IN2 = A2 = D27 = pin.GPIO2
GPIO_PC15 = ADC_IN0 = A0 = D22 = pin.GPIO0
GPIO_PC16 = D23 = pin.GPIO119
GPIO_PC18 = D24 = pin.GPIO114

GPIO_PC19 = ECSPI3_MOSI = D10 = MOSI = pin.ECSPI3_MOSI
GPIO_PC21 = ECSPI3_MISO = D9 = MISO = pin.ECSPI3_MISO
GPIO_PC22 = D25 = pin.GPIO27
GPIO_PC23 = ECSPI3_SCLK = D11 = SCLK = SCK = pin.ECSPI3_SCLK
GPIO_PC24 = ECSPI3_SS0 = D8 = SS0 = pin.ECSPI3_SS0
GPIO_PC26 = ECSPI3_SS1 = D7 = SS1 = pin.ECSPI3_SS1

GPIO_PC27 = I2C3_SDA = D0 = pin.I2C3_SDA
GPIO_PC28 = I2C3_SCL = D1 = pin.I2C3_SCL

GPIO_PC29 = D5 = pin.GPIO117
GPIO_PC31 = D6 = pin.GPIO118

GPIO_PC32 = LED_D6 = D12 = pin.GPIO115
GPIO_PC33 = LED_D5 = D13 = pin.GPIO116

# Board pwm channel = RPI PWM Channel = pin name
PWM_C7 = PWM1 = pin.GPIO115
PWM_C8 = PWM2 = pin.GPIO116

GPIO_PC35 = D19 = pin.GPIO121
GPIO_PC36 = D16 = pin.GPIO120
GPIO_PC37 = D26 = pin.GPIO26
GPIO_PC38 = D20 = pin.GPIO123
GPIO_PC40 = D21 = pin.GPIO124

# Mini header CN4
GPIO_PD9 = ADC_IN1 = A1 = pin.GPIO1
GPIO_PD4 = LED_D4 = PWM_C3 = pin.GPIO4
GPIO_PD17 = BUTTON2 = pin.GPIO129



"""PWMOut Class for NXP LPC4330"""

from greatfet import GreatFET

try:
    from microcontroller.pin import pwmOuts
except ImportError:
    raise RuntimeError("No PWM outputs defined for this board") from ImportError

# pylint: disable=unnecessary-pass
class PWMError(IOError):
    """Base class for PWM errors."""

    pass


# pylint: enable=unnecessary-pass
class PWMOut:
    """Pulse Width Modulation Output Class"""

    MAX_CYCLE_LEVEL = 1024

    def __init__(self, pin, *, frequency=750, duty_cycle=0, variable_frequency=False):
        """This class makes use of the GreatFET One's Pattern Generator to create a
        Simulated Pulse width modulation. The way that the Pattern Generator works is that
        takes a pattern in the form of bytes and will repeat the output. The trick to simulate
        PWM is to generate the correct byte pattern for the correct channel.

        Args:
            pin (Pin): CircuitPython Pin object to output to
            duty_cycle (int) : The fraction of each pulse which is high. 16-bit
            frequency (int) : target frequency in Hertz (32-bit)

        Returns:
            PWMOut: PWMOut object.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if `channel` or `pin` types are invalid.
            ValueError: if PWM channel does not exist.
        """
        self._gf = GreatFET()

        if variable_frequency:
            raise NotImplementedError("Variable Frequency is not currently supported.")

        self._pattern = None
        self._channel = None
        self._enable = False
        self._frequency = 500
        self._duty_cycle = 0
        self._open(pin, duty_cycle, frequency)

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.deinit()

    def _open(self, pin, duty=0, freq=500):
        self._channel = None
        for pwmpair in pwmOuts:
            if pwmpair[1] == pin:
                self._channel = pwmpair[0]

        self._pin = pin
        if self._channel is None:
            raise RuntimeError("No PWM channel found for this Pin")

        # set duty
        self.duty_cycle = duty

        # set frequency
        self.frequency = freq

        self._set_enabled(True)

    def deinit(self):
        """Deinit the GreatFET One PWM."""
        # pylint: disable=broad-except
        try:
            if self._channel is not None:
                # self.duty_cycle = 0
                self._set_enabled(False)
        except Exception as e:
            # due to a race condition for which I have not yet been
            # able to find the root cause, deinit() often fails
            # but it does not effect future usage of the pwm pin
            print(
                "warning: failed to deinitialize pwm pin {0} due to: {1}\n".format(
                    self._channel, type(e).__name__
                )
            )
        finally:
            self._pattern = None
            self._channel = None
        # pylint: enable=broad-except

    def _is_deinited(self):
        if self._pattern is None:
            raise ValueError(
                "Object has been deinitialize and can no longer "
                "be used. Create a new object."
            )

    # Mutable properties

    def _get_period(self):
        return 1.0 / self._get_frequency()

    def _set_period(self, period):
        """Get or set the PWM's output period in seconds.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not int or float.

        :type: int, float
        """
        if not isinstance(period, (int, float)):
            raise TypeError("Invalid period type, should be int or float.")

        self._set_frequency(1.0 / period)

    period = property(_get_period, _set_period)

    def _get_duty_cycle(self):
        """Get or set the PWM's output duty cycle as a ratio from 0.0 to 1.0.

        Raises:
           PWMError: if an I/O or OS error occurs.
           TypeError: if value type is not int or float.
           ValueError: if value is out of bounds of 0.0 to 1.0.

        :type: int, float
        """
        return self._duty_cycle

    def _set_duty_cycle(self, duty_cycle):
        if not isinstance(duty_cycle, (int, float)):
            raise TypeError("Invalid duty cycle type, should be int or float.")

        # convert from 16-bit
        if isinstance(duty_cycle, int):
            duty_cycle /= 65535.0
        if not 0.0 <= duty_cycle <= 1.0:
            raise ValueError("Invalid duty cycle value, should be between 0.0 and 1.0.")

        # Generate a pattern for 1024 samples of the duty cycle
        pattern = [(1 << self._channel)] * round(PWMOut.MAX_CYCLE_LEVEL * duty_cycle)
        pattern += [(0 << self._channel)] * round(
            PWMOut.MAX_CYCLE_LEVEL * (1.0 - duty_cycle)
        )

        self._pattern = pattern
        self._duty_cycle = duty_cycle
        if self._enable:
            self._set_enabled(True)

    duty_cycle = property(_get_duty_cycle, _set_duty_cycle)

    def _get_frequency(self):
        return self._frequency

    def _set_frequency(self, frequency):
        """Get or set the PWM's output frequency in Hertz.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not int or float.

        :type: int, float
        """
        if not isinstance(frequency, (int, float)):
            raise TypeError("Invalid frequency type, should be int or float.")

        # We are sending 1024 samples per second already
        self._gf.pattern_generator.set_sample_rate(frequency * len(self._pattern))
        self._frequency = frequency

    frequency = property(_get_frequency, _set_frequency)

    def _get_enabled(self):
        enabled = self._enable

        if enabled == "1":
            return True
        if enabled == "0":
            return False

        raise PWMError(None, 'Unknown enabled value: "%s"' % enabled)

    def _set_enabled(self, value):
        """Get or set the PWM's output enabled state.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not bool.

        :type: bool
        """
        if not isinstance(value, bool):
            raise TypeError("Invalid enabled type, should be string.")
        self._enable = value
        if self._gf:
            if self._enable:
                if self._pattern:
                    self._gf.pattern_generator.scan_out_pattern(self._pattern)
            else:
                self._gf.pattern_generator.stop()



"""Tegra T194 pin names"""
import atexit
import Jetson.GPIO as GPIO

GPIO.setmode(GPIO.TEGRA_SOC)
GPIO.setwarnings(False)  # shh!


class Pin:
    """Pins dont exist in CPython so...lets make our own!"""

    IN = 0
    OUT = 1
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2

    id = None
    _value = LOW
    _mode = IN

    def __init__(self, bcm_number):
        self.id = bcm_number

    def __repr__(self):
        return str(self.id)

    def __eq__(self, other):
        return self.id == other

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if mode is not None:
            if mode == self.IN:
                self._mode = self.IN
                GPIO.setup(self.id, GPIO.IN)
            elif mode == self.OUT:
                self._mode = self.OUT
                GPIO.setup(self.id, GPIO.OUT)
            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)
        if pull is not None:
            if self._mode != self.IN:
                raise RuntimeError("Cannot set pull resistor on output")
            if pull == self.PULL_UP:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            elif pull == self.PULL_DOWN:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            else:
                raise RuntimeError("Invalid pull for pin: %s" % self.id)

    def value(self, val=None):
        """Set or return the Pin Value"""
        if val is not None:
            if val == self.LOW:
                self._value = val
                GPIO.output(self.id, val)
                return None
            if val == self.HIGH:
                self._value = val
                GPIO.output(self.id, val)
                return None
            raise RuntimeError("Invalid value for pin")
        return GPIO.input(self.id)

    # pylint: disable=no-method-argument
    @atexit.register
    def cleanup():
        """Clean up pins"""
        print("Exiting... \nCleaning up pins")
        GPIO.cleanup()

    # pylint: enable=no-method-argument


# Cannot be used as GPIO
SDA = Pin("DP_AUX_CH3_N")
SCL = Pin("DP_AUX_CH3_P")
SDA_1 = Pin("GEN2_I2C_SDA")
SCL_1 = Pin("GEN2_I2C_SCL")

# Jetson Xavier only
Q06 = Pin("SOC_GPIO42")
AA03 = Pin("CAN0_DIN")
AA02 = Pin("CAN0_DOUT")
BB01 = Pin("CAN1_EN")
AA00 = Pin("CAN1_DOUT")
H07 = Pin("DAP2_SCLK")
I02 = Pin("DAP2_FS")
I01 = Pin("DAP2_DIN")
I00 = Pin("DAP2_DOUT")
BB00 = Pin("CAN1_STB")
H00 = Pin("SOC_GPIO12")
Q01 = Pin("SOC_GPIO21")
AA01 = Pin("CAN1_DIN")

# Jetson NX only
S04 = Pin("AUD_MCLK")
T05 = Pin("DAP5_SCLK")
Y00 = Pin("SPI3_SCK")
CC04 = Pin("TOUCH_CLK")
Y04 = Pin("SPI3_CS1_N")
Y03 = Pin("SPI3_CS0_N")
Y01 = Pin("SPI3_MISO")
Q05 = Pin("SOC_GPIO41")
Q06 = Pin("SOC_GPIO42")
U00 = Pin("DAP5_FS")
Y02 = Pin("SPI3_MOSI")
T07 = Pin("DAP5_DIN")
T06 = Pin("DAP5_DOUT")

# Clara AGX Xavier only
P04 = Pin("SOC_GPIO04")

# Shared
N01 = Pin("SOC_GPIO54")
R00 = Pin("SOC_GPIO44")
R04 = Pin("UART1_RTS")
R05 = Pin("UART1_CTS")
Z03 = Pin("SPI1_SCK")
Z04 = Pin("SPI1_MISO")
Z05 = Pin("SPI1_MOSI")
Z06 = Pin("SPI1_CS0_N")
Z07 = Pin("SPI1_CS1_N")

i2cPorts = (
    (8, SCL, SDA),
    (1, SCL_1, SDA_1),
)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, Z03, Z05, Z04),)



"""NodeMCU pin names"""

from adafruit_blinka.microcontroller.esp8266 import pin

D0 = pin.GPIO16
D1 = pin.GPIO5
D2 = pin.GPIO4
D3 = pin.GPIO0
D4 = pin.GPIO2
D5 = pin.GPIO14
D6 = pin.GPIO12
D7 = pin.GPIO13
D8 = pin.GPIO15
D9 = pin.GPIO3
D10 = pin.GPIO1

TX1 = D4
"""Transmit pin from second (transmit-only) UART """

CLK = D5
"""SPI clock pin"""
MISO = D6
"""SPI MISO (Master in, Slave out)"""
MOSI = D7
"""SPI MOSI (Master out, Slave in)"""

RX0 = D9
TX0 = D10

# GPIO0 and GPIO2 have built-in pull-ups on common ESP8266
# breakout boards making them suitable for I2C SDA and SCL



"""Generic Linux I2C class using PureIO's smbus class"""
import Adafruit_PureIO.smbus as smbus


class I2C:
    """I2C class"""

    MASTER = 0
    SLAVE = 1
    _baudrate = None
    _mode = None
    _i2c_bus = None

    # pylint: disable=unused-argument
    def __init__(self, bus_num, mode=MASTER, baudrate=None):
        if mode != self.MASTER:
            raise NotImplementedError("Only I2C Master supported!")
        _mode = self.MASTER

        # if baudrate != None:
        #    print("I2C frequency is not settable in python, ignoring!")

        try:
            self._i2c_bus = smbus.SMBus(bus_num)
        except FileNotFoundError:
            raise RuntimeError(
                "I2C Bus #%d not found, check if enabled in config!" % bus_num
            ) from RuntimeError

    # pylint: enable=unused-argument

    def scan(self):
        """Try to read a byte from each address, if you get an OSError
        it means the device isnt there"""
        found = []
        for addr in range(0, 0x80):
            try:
                self._i2c_bus.read_byte(addr)
            except OSError:
                continue
            found.append(addr)
        return found

    # pylint: disable=unused-argument
    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        """Write data from the buffer to an address"""
        if end is None:
            end = len(buffer)
        self._i2c_bus.write_bytes(address, buffer[start:end])

    def readfrom_into(self, address, buffer, *, start=0, end=None, stop=True):
        """Read data from an address and into the buffer"""
        if end is None:
            end = len(buffer)

        readin = self._i2c_bus.read_bytes(address, end - start)
        for i in range(end - start):
            buffer[i + start] = readin[i]

    # pylint: enable=unused-argument

    def writeto_then_readfrom(
        self,
        address,
        buffer_out,
        buffer_in,
        *,
        out_start=0,
        out_end=None,
        in_start=0,
        in_end=None,
        stop=False
    ):
        """Write data from buffer_out to an address and then
        read data from an address and into buffer_in
        """
        if out_end is None:
            out_end = len(buffer_out)
        if in_end is None:
            in_end = len(buffer_in)
        if stop:
            # To generate a stop in linux, do in two transactions
            self.writeto(address, buffer_out, start=out_start, end=out_end, stop=True)
            self.readfrom_into(address, buffer_in, start=in_start, end=in_end)
        else:
            # To generate without a stop, do in one block transaction
            readin = self._i2c_bus.read_i2c_block_data(
                address, buffer_out[out_start:out_end], in_end - in_start
            )
            for i in range(in_end - in_start):
                buffer_in[i + in_start] = readin[i]



"""Pin definitions for Jetson NX."""

from adafruit_blinka.microcontroller.tegra.t194 import pin

SDA = pin.SDA
SCL = pin.SCL
SDA_1 = pin.SDA_1
SCL_1 = pin.SCL_1

D4 = pin.S04
D5 = pin.Q05
D6 = pin.Q06
D7 = pin.Z07
D8 = pin.Z06
D9 = pin.Z04
D10 = pin.Z05
D11 = pin.Z03
D12 = pin.R00
D13 = pin.N01
D16 = pin.R05
D17 = pin.R04
D18 = pin.T05
D19 = pin.U00
D20 = pin.T07
D21 = pin.T06
D22 = pin.CC04
D23 = pin.Y04
D24 = pin.Y03
D25 = pin.Y01
D26 = pin.Y02
D27 = pin.Y00

CE1 = D7
CE0 = D8
MISO = D9
MOSI = D10
SCLK = D11
SCK = D11



"""MediaTek MT8167 pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

# All pins
GPIO52 = Pin(52)  # SDA1      (pin 3)
GPIO53 = Pin(53)  # SCL1      (pin 5)
GPIO22 = Pin(22)  # EINT22    (pin 7)
GPIO63 = Pin(63)  # UTXD0     (pin 8)
GPIO62 = Pin(62)  # URXD0     (pin 10)
GPIO9 = Pin(9)  # EINT9     (pin 11)
GPIO36 = Pin(36)  # MRG_CLK   (pin 12)
GPIO10 = Pin(10)  # EINT10    (pin 13)
GPIO11 = Pin(11)  # EINT11    (pin 15)
GPIO0 = Pin(0)  # EINT0     (pin 16)
GPIO1 = Pin(1)  # EINT1     (pin 18)
GPIO4 = Pin(4)  # EINT4     (pin 19)
GPIO3 = Pin(3)  # EINT3     (pin 21)
GPIO7 = Pin(7)  # EINT7     (pin 22)
GPIO6 = Pin(6)  # EINT6     (pin 23)
GPIO5 = Pin(5)  # EINT5     (pin 24)
GPIO8 = Pin(8)  # EINT8     (pin 26)
GPIO60 = Pin(60)  # SDA2      (pin 27)
GPIO61 = Pin(61)  # SCL2      (pin 28)
GPIO65 = Pin(65)  # UTXD1     (pin 29)
GPIO64 = Pin(64)  # URXD1     (pin 31)
GPIO12 = Pin(12)  # EINT12    (pin 32)
GPIO25 = Pin(25)  # EINT25    (pin 33)
GPIO37 = Pin(37)  # MRG_SYNC  (pin 35)
GPIO13 = Pin(13)  # EINT13    (pin 36)
GPIO45 = Pin(45)  # JTCLK     (pin 37)
GPIO38 = Pin(38)  # MRG_DI    (pin 38)
GPIO39 = Pin(39)  # MRG_DO    (pin 40)

# Aliases
PWM_A = GPIO25  # EINT12  (pin 32)
PWM_B = GPIO11  # EINT25  (pin 33)
PWM_C = GPIO12  # EINT11  (pin 15)

I2C1_SDA = GPIO52  # SDA1  (pin 3)
I2C1_SCL = GPIO53  # SCL1  (pin 5)

I2C2_SDA = GPIO60  # SDA2  (pin 27)
I2C2_SCL = GPIO61  # SCL2  (pin 28)

SPI_MO = GPIO4  # EINT4  (pin 19)
SPI_MI = GPIO3  # EINT3  (pin 21)
SPI_CLK = GPIO6  # EINT6  (pin 23)
SPI_CSB = GPIO5  # EINT5  (pin 24)

# SysFS pwm outputs, pwm channel and pin in first tuple
pwmOuts = (
    ((0, 0), PWM_A),
    ((0, 1), PWM_B),
    ((0, 2), PWM_C),
)

# ordered as i2cId, sclId, sdaId
i2cPorts = (
    (3, I2C1_SCL, I2C1_SDA),
    (0, I2C2_SCL, I2C2_SDA),
)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, SPI_CLK, SPI_MO, SPI_MI),)



"""Custom PWMOut Wrapper for Rpi.GPIO PWM Class"""
import RPi.GPIO as GPIO

GPIO.setmode(GPIO.BCM)  # Use BCM pins D4 = GPIO #4
GPIO.setwarnings(False)  # shh!


# pylint: disable=unnecessary-pass
class PWMError(IOError):
    """Base class for PWM errors."""

    pass


# pylint: enable=unnecessary-pass


class PWMOut:
    """Pulse Width Modulation Output Class"""

    def __init__(self, pin, *, frequency=500, duty_cycle=0, variable_frequency=False):
        self._pwmpin = None
        self._period = 0
        self._open(pin, duty_cycle, frequency, variable_frequency)

    def __del__(self):
        self.deinit()

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.deinit()

    def _open(self, pin, duty=0, freq=500, variable_frequency=False):
        self._pin = pin
        GPIO.setup(pin.id, GPIO.OUT)
        self._pwmpin = GPIO.PWM(pin.id, freq)

        if variable_frequency:
            print("Variable Frequency is not supported, continuing without it...")

        # set frequency
        self.frequency = freq
        # set duty
        self.duty_cycle = duty

        self.enabled = True

    def deinit(self):
        """Deinit the PWM."""
        self._pwmpin.stop()
        GPIO.cleanup(self._pin.id)
        self._pwmpin = None

    def _is_deinited(self):
        if self._pwmpin is None:
            raise ValueError(
                "Object has been deinitialize and can no longer "
                "be used. Create a new object."
            )

    @property
    def period(self):
        """Get or set the PWM's output period in seconds.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not int or float.

        :type: int, float
        """
        return 1.0 / self.frequency

    @period.setter
    def period(self, period):
        if not isinstance(period, (int, float)):
            raise TypeError("Invalid period type, should be int or float.")

        self.frequency = 1.0 / period

    @property
    def duty_cycle(self):
        """Get or set the PWM's output duty cycle which is the fraction of
        each pulse which is high. 16-bit

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not int or float.
            ValueError: if value is out of bounds of 0.0 to 1.0.

        :type: int, float
        """
        return int(self._duty_cycle * 65535)

    @duty_cycle.setter
    def duty_cycle(self, duty_cycle):
        if not isinstance(duty_cycle, (int, float)):
            raise TypeError("Invalid duty cycle type, should be int or float.")

        if not 0 <= duty_cycle <= 65535:
            raise ValueError("Invalid duty cycle value, should be between 0 and 65535")

        # convert from 16-bit
        duty_cycle /= 65535.0

        self._duty_cycle = duty_cycle
        self._pwmpin.ChangeDutyCycle(round(self._duty_cycle * 100))

    @property
    def frequency(self):
        """Get or set the PWM's output frequency in Hertz.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not int or float.

        :type: int, float
        """

        return self._frequency

    @frequency.setter
    def frequency(self, frequency):
        if not isinstance(frequency, (int, float)):
            raise TypeError("Invalid frequency type, should be int or float.")

        self._pwmpin.ChangeFrequency(round(frequency))
        self._frequency = frequency

    @property
    def enabled(self):
        """Get or set the PWM's output enabled state.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not bool.

        :type: bool
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if not isinstance(value, bool):
            raise TypeError("Invalid enabled type, should be string.")

        if value:
            self._pwmpin.start(round(self._duty_cycle * 100))
        else:
            self._pwmpin.stop()

        self._enabled = value

    # String representation
    def __str__(self):
        return "pin %s (freq=%f Hz, duty_cycle=%f%%)" % (
            self._pin,
            self.frequency,
            self.duty_cycle,
        )



"""Pin definitions for Udoo x86 Ultra
(should work for Ultra and Ultra II)

There are 2 naming systems.  A Digital Pin system which includes
the arduino chip (Leonardo or 101).  The Braswell #s start at 16 and
is documented in the diagram and text for linux version later than 4.15 in:
https://www.udoo.org/docs-x86II/Introduction/Introduction.html
The other is based on the hardware manual:
https://udoo.org/download/files/UDOO_X86/Doc/UDOO_X86II_MANUAL.pdf

This will use the D system based on the diagram in the user guide

i2c use i2c(board.I2C0_SCL, board_I2C0_SDA) or i2c(board.I2C1_SCL, board_I2C1_SDA)
for the i2c(board.SCL, board.SCL) in the examples

UART use pyserial not busio
"""

from adafruit_blinka.microcontroller.pentium.n3710 import pin

# Connector CN15
D16 = pin.UART1_RTS
D17 = pin.UART1_CTS
D18 = pin.UART1_TXD
D19 = pin.UART1_RXD
D20 = pin.UART2_RTS
D21 = pin.UART2_CTS
D22 = pin.UART2_TXD
D23 = pin.UART2_RXD

# Connector CN13 LPC interface
D24 = pin.GPIO_275
D25 = pin.GPIO_280
D26 = pin.GPIO_273
D27 = pin.GPIO_278
D28 = pin.GPIO_276
D29 = pin.GPIO_279
D30 = pin.GPIO_307

# Connector CN14
D34 = pin.I2C0_SCL
D35 = pin.I2C0_SDA

D36 = pin.GPIO_492
D37 = pin.GPIO_490

D38 = pin.I2C1_SCL
D39 = pin.I2C1_SDA

# Connector CN12 SDIO SD/MMC interfaces
D40 = pin.GPIO_358
D41 = pin.GPIO_243
D42 = pin.GPIO_249
D43 = pin.GPIO_246
D44 = pin.GPIO_253
D45 = pin.GPIO_250
D46 = pin.GPIO_247

# aliases
UART1_RX = D19
UART1_TX = D18

UART2_RX = D23
UART2_TX = D22

I2C0_SCL = D34  # labeled on diagram as I2C1, hardware manual port 0
I2C0_SDA = D35  # i2cdetect-l lists it as i2c-0

I2C1_SCL = D38  # Labeled on diagram as I2C2, hardware manual port 5
I2C1_SCL = D39  # i2cdetect lists it as i2c-1



"""Configuration of testing fixtures depending on the board layout"""
from adafruit_blinka import agnostic

import board

if agnostic.board == "feather_m0_express":
    default_pin = board.D5
    led_pin = board.D13
    led_hardwired = True
    led_inverted = False
elif agnostic.board == "feather_huzzah":
    default_pin = board.GPIO4
    led_pin = board.GPIO0  # red led
    led_hardwired = True
    led_inverted = True
elif agnostic.board == "pyboard":
    default_pin = board.X1
    led_pin = board.LED_BLUE
    led_hardwired = True
    led_inverted = False
    uartTxId = "B6"
    uartRXId = "B7"
else:
    raise NotImplementedError("Board not supported")



"""NXP LPC4330 pin names"""
try:
    from greatfet import GreatFET
    from greatfet.interfaces.adc import ADC

    gf = GreatFET()
except ModuleNotFoundError:
    raise RuntimeError(
        "Unable to create GreatFET object. Make sure library is "
        "installed and the device is connected."
    ) from ModuleNotFoundError


class Pin:
    """A basic Pin class for the NXP LPC4330 that
    acts as a wrapper for the GreatFET api.
    """

    # pin modes
    OUT = gf.gpio.DIRECTION_OUT
    IN = gf.gpio.DIRECTION_IN
    ADC = 2
    DAC = 3

    # pin values
    LOW = 0
    HIGH = 1

    def __init__(self, pin_id=None):
        self.id = pin_id
        self._mode = None
        self._pin = None

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if self.id is None:
            raise RuntimeError("Can not init a None type pin.")
        if pull is not None:
            raise NotImplementedError("Internal pullups and pulldowns not supported")
        if mode in (Pin.IN, Pin.OUT):
            if self.id not in gf.GPIO_MAPPINGS:
                raise ValueError("Pin does not have GPIO capabilities")
            self._pin = gf.gpio.get_pin(self.id)
            self._pin.set_direction(mode)
        elif mode == Pin.ADC:
            # ADC only available on these pins
            if self.id not in gf.ADC_MAPPINGS:
                raise ValueError("Pin does not have ADC capabilities")
            self._pin = ADC(gf, self.id)
        elif mode == Pin.DAC:
            # DAC only available on these pins
            if self.id != "J2_P5":
                raise ValueError("Pin does not have DAC capabilities")
            self._pin = gf.apis.dac
            self._pin.initialize()
        else:
            raise ValueError("Incorrect pin mode: {}".format(mode))
        self._mode = mode

    def value(self, val=None):
        """Set or return the Pin Value"""
        # Digital In / Out
        if self._mode in (Pin.IN, Pin.OUT):
            # digital read
            if val is None:
                return self._pin.get_state()
            # digital write
            if val in (Pin.LOW, Pin.HIGH):
                self._pin.set_state(val)
                return None
            # nope
            raise ValueError("Invalid value for pin.")
        # Analog In
        if self._mode == Pin.ADC:
            if val is None:
                # Read ADC here
                return self._pin.read_samples()[0]
            # read only
            raise AttributeError("'AnalogIn' object has no attribute 'value'")
        # Analog Out
        if self._mode == Pin.DAC:
            if val is None:
                # write only
                raise AttributeError("unreadable attribute")
            # Set DAC Here
            self._pin.set_value(int(val))
            return None
        raise RuntimeError(
            "No action for mode {} with value {}".format(self._mode, val)
        )


# create pin instances for each pin
# J1 Header Pins
J1_P3 = Pin("J1_P3")
J1_P4 = Pin("J1_P4")
J1_P5 = Pin("J1_P5")
J1_P6 = Pin("J1_P6")
J1_P7 = Pin("J1_P7")
J1_P8 = Pin("J1_P8")
J1_P9 = Pin("J1_P9")
J1_P10 = Pin("J1_P10")
J1_P12 = Pin("J1_P12")
J1_P13 = Pin("J1_P13")
J1_P14 = Pin("J1_P14")
J1_P15 = Pin("J1_P15")
J1_P16 = Pin("J1_P16")
J1_P17 = Pin("J1_P17")
J1_P18 = Pin("J1_P18")
J1_P19 = Pin("J1_P19")
J1_P20 = Pin("J1_P20")
J1_P21 = Pin("J1_P21")
J1_P22 = Pin("J1_P22")
J1_P23 = Pin("J1_P23")
J1_P24 = Pin("J1_P24")
J1_P25 = Pin("J1_P25")
J1_P26 = Pin("J1_P26")
J1_P27 = Pin("J1_P27")
J1_P28 = Pin("J1_P28")
J1_P29 = Pin("J1_P29")
J1_P30 = Pin("J1_P30")
J1_P31 = Pin("J1_P31")
J1_P32 = Pin("J1_P32")
J1_P33 = Pin("J1_P33")
J1_P34 = Pin("J1_P34")
J1_P35 = Pin("J1_P35")
J1_P37 = Pin("J1_P37")
J1_P39 = Pin("J1_P39")  # MOSI
J1_P40 = Pin("J1_P40")  # MISO


# J2 Header Pins
J2_P3 = Pin("J2_P3")
J2_P4 = Pin("J2_P4")
J2_P5 = Pin("J2_P5")  # ADC, ADC, DAC
J2_P6 = Pin("J2_P6")
J2_P7 = Pin("J2_P7")
J2_P8 = Pin("J2_P8")
J2_P9 = Pin("J2_P9")  # ADC, GPIO
J2_P10 = Pin("J2_P10")
J2_P13 = Pin("J2_P13")
J2_P14 = Pin("J2_P14")
J2_P15 = Pin("J2_P15")
J2_P16 = Pin("J2_P16")  # GPIO, ADC
J2_P18 = Pin("J2_P18")
J2_P19 = Pin("J2_P19")
J2_P20 = Pin("J2_P20")
J2_P22 = Pin("J2_P22")
J2_P23 = Pin("J2_P23")
J2_P24 = Pin("J2_P24")
J2_P25 = Pin("J2_P25")
J2_P27 = Pin("J2_P27")
J2_P28 = Pin("J2_P28")
J2_P29 = Pin("J2_P29")
J2_P30 = Pin("J2_P30")
J2_P31 = Pin("J2_P31")
J2_P33 = Pin("J2_P33")
J2_P34 = Pin("J2_P34")
J2_P35 = Pin("J2_P35")
J2_P36 = Pin("J2_P36")
J2_P37 = Pin("J2_P37")
J2_P38 = Pin("J2_P38")

# Bonus Row Pins
J7_P2 = Pin("J7_P2")
J7_P3 = Pin("J7_P3")
J7_P4 = Pin("J7_P4")  # ADC, ADC
J7_P5 = Pin("J7_P5")  # ADC, ADC
J7_P6 = Pin("J7_P6")
J7_P7 = Pin("J7_P7")
J7_P8 = Pin("J7_P8")
J7_P13 = Pin("J7_P13")
J7_P14 = Pin("J7_P14")
J7_P15 = Pin("J7_P15")
J7_P16 = Pin("J7_P16")
J7_P17 = Pin("J7_P17")
J7_P18 = Pin("J7_P18")

SCL = Pin()
SDA = Pin()

SCK = Pin()
MOSI = J1_P39
MISO = J1_P40

TX = J1_P33
RX = J1_P34

# ordered as uartId, txId, rxId
uartPorts = ((0, TX, RX),)

# pwm outputs: pwm channel and pin
pwmOuts = (
    (0, J1_P4),
    (1, J1_P6),
    (2, J1_P28),
    (3, J1_P30),
    (4, J2_P36),
    (5, J2_P34),
    (6, J2_P33),
    (7, J1_P34),
    (8, J2_P9),
    (9, J1_P6),
    (10, J1_P25),
    (11, J1_P32),
    (12, J1_P31),
    (13, J2_P3),
    (14, J1_P3),
    (15, J1_P5),
)
